from abc import ABCMeta, abstractmethod


class BaseFileServer(metaclass=ABCMeta):
    name = 'base'

    def __init__(self, filename=None, **type_args):
        self.filename = filename
        self.type_args = type_args

    @abstractmethod
    def serve(self, request):
        raise NotImplemented



from logging import getLogger

from .base import BaseNotifier

logger = getLogger(__name__)


class NullNotifier(BaseNotifier):
    name = 'none'

    def notify_start(self):
        logger.debug('Processing started.')

    def notify_end(self, file_url):
        logger.debug('Processing ended.')

    def notify_error(self):
        logger.debug('Processing failed.')



import pytest
from mock import patch, MagicMock
from splitgill.indexing.utils import DOC_TYPE

from ckanext.versioned_datastore.lib.basic_query.utils import format_facets, get_fields


class TestBasicQueryUtils(object):
    def test_format_facets(self):
        # first, check it can deal with an empty aggregation result
        assert format_facets({}) == {}

        facets = format_facets(
            {
                'facet1': {
                    'sum_other_doc_count': 901,
                    'doc_count_error_upper_bound': 12,
                    'buckets': [
                        {'key': 'value1', 'doc_count': 43},
                        {'key': 'value2', 'doc_count': 243},
                        {'key': 'value3', 'doc_count': 543},
                        {'key': 'value4', 'doc_count': 143},
                        {'key': 'value5', 'doc_count': 743},
                    ],
                },
                'facet2': {
                    'sum_other_doc_count': 0,
                    'doc_count_error_upper_bound': 0,
                    'buckets': [
                        {'key': 'value1', 'doc_count': 6},
                    ],
                },
            }
        )

        assert len(facets) == 2
        assert facets['facet1']['details']['sum_other_doc_count'] == 901
        assert facets['facet1']['details']['doc_count_error_upper_bound'] == 12
        assert len(facets['facet1']['values']) == 5
        assert facets['facet1']['values']['value1'] == 43
        assert facets['facet1']['values']['value2'] == 243
        assert facets['facet1']['values']['value3'] == 543
        assert facets['facet1']['values']['value4'] == 143
        assert facets['facet1']['values']['value5'] == 743

        assert facets['facet2']['details']['sum_other_doc_count'] == 0
        assert facets['facet2']['details']['doc_count_error_upper_bound'] == 0
        assert len(facets['facet2']['values']) == 1
        assert facets['facet2']['values']['value1'] == 6

    @pytest.mark.filterwarnings('ignore::sqlalchemy.exc.SADeprecationWarning')
    @pytest.mark.ckan_config('ckan.plugins', 'versioned_datastore')
    @pytest.mark.usefixtures('with_versioned_datastore_tables', 'with_plugins')
    def test_get_fields(self):
        mock_mapping = {
            u"beans-index": {
                u"mappings": {
                    DOC_TYPE: {
                        u"properties": {
                            u"data": {
                                u"properties": {
                                    u"_id": {'type': 'long'},
                                    u"field1": {
                                        u"type": u"keyword",
                                    },
                                    u"field2": {
                                        u"type": u"date",
                                    },
                                }
                            }
                        }
                    }
                }
            }
        }

        mapping_mock_function = MagicMock(return_value=mock_mapping)
        prefix_mock = lambda name: f'beans-{name}'
        client_mock = MagicMock(indices=MagicMock(get_mapping=mapping_mock_function))
        search_helper_mock = MagicMock()
        es_response = [
            MagicMock(hits=MagicMock(total=4)),
            MagicMock(hits=MagicMock(total=10)),
        ]
        multisearch_mock = MagicMock()
        multisearch_mock.configure_mock(
            add=MagicMock(return_value=multisearch_mock),
            execute=MagicMock(return_value=es_response),
        )
        multisearch_class_mock = MagicMock(return_value=multisearch_mock)

        with patch(
            'ckanext.versioned_datastore.lib.basic_query.utils.prefix_resource',
            new=prefix_mock,
        ), patch(
            'ckanext.versioned_datastore.lib.common.ES_CLIENT', new=client_mock
        ), patch(
            'ckanext.versioned_datastore.lib.common.SEARCH_HELPER',
            new=search_helper_mock,
        ), patch(
            'ckanext.versioned_datastore.lib.basic_query.utils.MultiSearch',
            new=multisearch_class_mock,
        ):
            mapping, fields = get_fields('index')
            assert mapping == mock_mapping['beans-index']
            assert len(fields) == 3
            # the first field should always be the _id field and it should always be an integer type
            assert fields[0] == {'id': '_id', 'type': 'integer'}
            assert {'id': 'field1', 'type': 'string'} in fields
            assert {'id': 'field2', 'type': 'string'} in fields



from abc import ABCMeta, abstractmethod
from jinja2 import Template

from ckan.plugins import toolkit, PluginImplementations
from ....interfaces import IVersionedDatastoreDownloads
from ....model.downloads import DownloadRequest


class BaseNotifier(metaclass=ABCMeta):
    name = 'base'

    default_start_text = '''
    Your download on {{ site_name }} has started processing.
    The status of your download can be viewed here: {{ status_page }}
    '''.strip()

    default_start_html = '''
    <p>Your download on <a href="{{ site_url }}">{{ site_name }}</a> has started processing.</p>
    <p>The status of your download can be viewed <a href="{{ status_page }}">here</a>.</p>
    '''.strip()

    default_end_text = '''
    The link to the resource data you requested on {{ site_url }} is available at {{ download_url }}.
    '''.strip()

    default_end_html = '''
    <p>The link to the resource data you requested on <a href="{{ site_url }}">{{ site_name }}</a>
    is available <a href="{{ download_url }}">here</a>.</p>
    '''.strip()

    default_error_text = '''
    Your download on {{ site_name }} has encountered an error and has stopped processing.
    More details can be viewed at: {{ status_page }}.
    Please try again later and contact us at {{ contact_email }} if the problem persists.
    '''.strip()

    default_error_html = '''
    <p>Your download on <a href="{{ site_url }}">{{ site_name }}</a> has encountered an error and
    has stopped processing.</p>
    <p>More details can be viewed <a href="{{ status_page }}">here</a>.</p>
    <p>Please try again later and contact us at <a href="mailto:{{ contact_email }}">{{ contact_email }}</a> if the problem persists.</p>
    '''.strip()

    def __init__(self, request, **type_args):
        self._request = request
        self.type_args = type_args

    @property
    def request(self):
        # this refreshes the joins if necessary
        if self._request is None:
            return
        try:
            self._request.core_record
        except:
            self._request = DownloadRequest.get(self._request.id)
        return self._request

    def template_context(self, file_url=None):
        context = {
            'site_url': toolkit.config.get('ckan.site_url'),
            'site_name': toolkit.config.get(
                'ckan.site_name', toolkit.config.get('ckan.site_url')
            ),
            'status_page': toolkit.url_for(
                'datastore_status.download_status',
                download_id=self.request.id,
                qualified=True,
            ),
            'download_url': file_url,
            'contact_email': toolkit.config.get('smtp.mail_from'),
        }
        for plugin in PluginImplementations(IVersionedDatastoreDownloads):
            context = plugin.download_modify_notifier_template_context(
                self.request, context
            )
        return context

    def _get_text(self, templates, interface_method_name, file_url=None):
        for plugin in PluginImplementations(IVersionedDatastoreDownloads):
            modify_templates = getattr(plugin, interface_method_name)
            templates = modify_templates(*templates)
        context = self.template_context(file_url)
        body, body_html = (
            Template(template).render(**context) for template in templates
        )
        return body, body_html

    def start_text(self):
        templates = (self.default_start_text, self.default_start_html)
        return self._get_text(templates, 'download_modify_notifier_start_templates')

    def end_text(self, file_url):
        templates = (self.default_end_text, self.default_end_html)
        return self._get_text(
            templates, 'download_modify_notifier_end_templates', file_url
        )

    def error_text(self):
        templates = (self.default_error_text, self.default_error_html)
        return self._get_text(templates, 'download_modify_notifier_error_templates')

    @abstractmethod
    def notify_start(self):
        raise NotImplemented

    @abstractmethod
    def notify_end(self, file_url):
        raise NotImplemented

    @abstractmethod
    def notify_error(self):
        raise NotImplemented

    @classmethod
    def validate_args(cls, type_args):
        return True



import os.path
from collections import defaultdict

import fastavro
import hashlib
import json
import os
import shutil
import tempfile
import zipfile
from datetime import datetime as dt
from elasticsearch_dsl import Search
from functools import partial
from glob import iglob
from splitgill.indexing.utils import get_elasticsearch_client
from splitgill.search import create_version_query

from ckan.lib import uploader
from ckan.plugins import toolkit, PluginImplementations
from .loaders import (
    get_derivative_generator,
    get_file_server,
    get_notifier,
    get_transformation,
)
from .query import Query
from .utils import get_schemas, calculate_field_counts, filter_data_fields, get_fields
from .. import common
from ..datastore_utils import prefix_resource
from ...interfaces import IVersionedDatastoreDownloads
from ...logic.actions.meta.arg_objects import DerivativeArgs
from ...model.downloads import CoreFileRecord, DownloadRequest
from ...model.downloads import DerivativeFileRecord


class DownloadRunManager:
    download_dir = toolkit.config.get('ckanext.versioned_datastore.download_dir')
    if download_dir is None:
        raise Exception('ckanext.versioned_datastore.download_dir must be set.')
    core_dir = os.path.join(download_dir, 'core')
    custom_dir = toolkit.config.get(
        'ckanext.versioned_datastore.custom_dir', os.path.join(download_dir, 'custom')
    )

    def __init__(self, query_args, derivative_args, server_args, notifier_args):
        # allow plugins to make changes to the args
        for plugin in PluginImplementations(IVersionedDatastoreDownloads):
            (
                query_args,
                derivative_args,
                server_args,
                notifier_args,
            ) = plugin.download_before_run(
                query_args, derivative_args, server_args, notifier_args
            )

        if derivative_args.format == 'raw':
            # if we're serving raw files then we're not doing any transformations
            derivative_args.separate_files = True
            derivative_args.ignore_empty_fields = False
            derivative_args.transform = {}
            # we also can't use a query
            query_args.query = {}
            query_args.slug_or_doi = None
            # resource versions will always be the latest version
            query_args.resource_ids = query_args.resource_ids or list(
                query_args.resource_ids_and_versions.keys()
            )
            query_args.resource_ids_and_versions = None

        self.allow_non_datastore = (
            derivative_args.format == 'raw'
            and derivative_args.format_args.get('allow_non_datastore', False)
        )
        self.query = Query.from_query_args(
            query_args, allow_non_datastore=self.allow_non_datastore
        )
        self.derivative_options = derivative_args
        for field, default_value in DerivativeArgs.defaults.items():
            if getattr(self.derivative_options, field) is None:
                setattr(self.derivative_options, field, default_value)
        self.server = get_file_server(
            server_args.type,
            filename=server_args.custom_filename,
            **server_args.type_args,
        )

        # initialise core and derivative records
        self.core_record, self.derivative_record = self.check_for_records()

        # initialises a log entry in the database
        self.request = DownloadRequest(
            core_id=self.core_record.id,
            derivative_id=self.derivative_record.id,
            server_args={f: getattr(server_args, f) for f in server_args.fields},
        )
        self.request.save()

        self.notifier = get_notifier(
            notifier_args.type, request=self.request, **notifier_args.type_args
        )

        self._temp = []

        for plugin in PluginImplementations(IVersionedDatastoreDownloads):
            plugin.download_after_init(self.request)

    @property
    def derivative_hash(self):
        """
        Unique hash for the derivative file options.
        """
        file_options = {
            f: getattr(self.derivative_options, f)
            for f in self.derivative_options.fields
        }
        file_options_hash = hashlib.sha1(json.dumps(file_options).encode('utf-8'))
        return file_options_hash.hexdigest()

    @property
    def hash(self):
        """
        Unique hash for this download request, identified by the query and the
        derivative options.
        """
        to_hash = [self.query.record_hash, self.derivative_hash]
        download_hash = hashlib.sha1('|'.join(to_hash).encode('utf-8'))
        return download_hash.hexdigest()

    @property
    def core_folder_path(self):
        """
        Location of the core files for this query.
        """
        return os.path.join(self.core_dir, self.query.hash)

    def run(self):
        """
        Run the download process.
        """
        try:
            # refresh the db objects because they were probably retrieved in a
            # different session (the __init__ is run by the main ckan process,
            # this is run by the download worker)
            self.request = DownloadRequest.get(self.request.id)
            self.core_record = CoreFileRecord.get(self.core_record.id)
            self.derivative_record = DerivativeFileRecord.get(self.derivative_record.id)

            self.notifier.notify_start()
            self.request.update_status(DownloadRequest.state_initial)

            # generate core file if needed
            # technically we don't need core files if the format is 'raw', but we'll
            # generate them anyway for consistency/avoidance of thousands of ifs, and
            # also so we have some field metadata
            self.generate_core()

            # generate derivative file if needed
            self.generate_derivative()

            # finish up
            self.request.update_status(DownloadRequest.state_complete)
            url = self.server.serve(self.request)
            self.notifier.notify_end(url)
        except Exception as e:
            current_state = self.request.state
            self.request.update_status(
                DownloadRequest.state_failed,
                f'{current_state}, {e.__class__.__name__}: {str(e)}',
            )
            self.notifier.notify_error()
            raise e
        finally:
            for t in self._temp:
                try:
                    shutil.rmtree(t)
                except FileNotFoundError:
                    pass
            for plugin in PluginImplementations(IVersionedDatastoreDownloads):
                plugin.download_after_run(self.request)

    def check_for_records(self):
        """
        Check if relevant files and records already exist and returns the records if
        they exist.

        :return: tuple of core_record (or None) and derivative_record (or None)
        """
        # check the download dir exists
        if not os.path.exists(self.download_dir):
            os.mkdir(self.download_dir)

        # also check the core dir exists
        if not os.path.exists(self.core_dir):
            os.mkdir(self.core_dir)

        # initialise empty variables
        core_record = None
        derivative_record = None

        # search for derivative first
        fp = os.path.join(self.download_dir, f'{self.hash}.zip')
        if os.path.exists(fp):
            existing_file = fp
        else:
            # check old-style names with the request id in front as well
            fn = f'*_{self.hash}.zip'
            existing_file = next(iglob(os.path.join(self.download_dir, fn)), None)
        if existing_file is not None:
            # could return multiple options, sorted by most recent first
            possible_records = DerivativeFileRecord.get_by_filepath(existing_file)
            if possible_records:
                # use the most recent one
                derivative_record = possible_records[0]
                core_record = derivative_record.core_record

        # if the core record hasn't been found by searching for the derivative, try and find it now
        if core_record is None:
            if os.path.exists(self.core_folder_path):
                # could return multiple options, sorted by most recent first
                possible_records = CoreFileRecord.get_by_hash(
                    self.query.hash, self.query.resource_hash
                )
                if possible_records:
                    # use the most recent one
                    core_record = possible_records[0]

        if core_record is None:
            core_record = CoreFileRecord(
                query_hash=self.query.hash,
                query=self.query.query,
                query_version=self.query.query_version,
                resource_ids_and_versions=self.query.resource_ids_and_versions,
                resource_hash=self.query.resource_hash,
            )
            core_record.save()

        if derivative_record is None:
            derivative_record = DerivativeFileRecord(
                core_id=core_record.id,
                download_hash=self.hash,
                format=self.derivative_options.format,
                options={
                    f: getattr(self.derivative_options, f)
                    for f in self.derivative_options.fields
                    if f != 'format'
                },
            )
            derivative_record.save()

        return core_record, derivative_record

    def generate_core(self):
        """
        Generates and loads core files.

        This method will add new resources to existing core records
        with identical filters, reducing data duplication.
        :return: the core record
        """
        if not os.path.exists(self.core_folder_path):
            os.makedirs(self.core_folder_path)

        # check if there are new resources to generate
        existing_files = os.listdir(self.core_folder_path)
        resources_to_generate = {
            rid: v
            for rid, v in self.query.resource_ids_and_versions.items()
            if f'{rid}_{v}.avro' not in existing_files
            and v != common.NON_DATASTORE_VERSION
        }

        resource_totals = {k: 0 for k in self.core_record.resource_ids_and_versions}
        field_counts = {k: None for k in self.core_record.resource_ids_and_versions}

        # get info for resources that weren't just generated first, so we know if they
        # need regenerating
        existing_resources = [
            k
            for k, v in self.query.resource_ids_and_versions.items()
            if k not in resources_to_generate and v != common.NON_DATASTORE_VERSION
        ]
        for resource_id in existing_resources:
            # find a matching core record
            record = CoreFileRecord.find_resource(
                self.query.hash,
                resource_id,
                self.query.resource_ids_and_versions[resource_id],
            )
            if record:
                resource_totals[resource_id] = record.resource_totals[resource_id]
                field_counts[resource_id] = record.field_counts[resource_id]
            else:
                # if there's no record we should regenerate the avro file
                resource_version = self.query.resource_ids_and_versions[resource_id]
                core_file_path = os.path.join(
                    self.core_folder_path, f'{resource_id}_{resource_version}.avro'
                )
                os.remove(core_file_path)
                resources_to_generate[resource_id] = resource_version

        if len(resources_to_generate) > 0:
            es_client = get_elasticsearch_client(
                common.CONFIG,
                sniff_on_start=True,
                sniffer_timeout=60,
                sniff_on_connection_fail=True,
                sniff_timeout=10,
                http_compress=False,
                timeout=30,
            )

            schemas = get_schemas(self.query)

            for resource_id, version in resources_to_generate.items():
                self.request.update_status(DownloadRequest.state_core_gen, resource_id)
                resource_totals[resource_id] = 0
                field_counts[resource_id] = calculate_field_counts(
                    self.query, es_client, resource_id, version
                )

                search = (
                    Search.from_dict(self.query.translate().to_dict())
                    .index(prefix_resource(resource_id))
                    .using(es_client)
                    .filter(create_version_query(version))
                )

                fn = f'{resource_id}_{version}.avro'
                fp = os.path.join(self.core_folder_path, fn)

                codec_kwargs = dict(codec='bzip2', codec_compression_level=9)
                chunk_size = 10000
                with open(fp, 'wb') as f:
                    fastavro.writer(f, schemas[resource_id], [], **codec_kwargs)

                def _flush(record_block):
                    with open(fp, 'a+b') as outfile:
                        fastavro.writer(outfile, None, record_block, **codec_kwargs)

                chunk = []
                for hit in search.scan():
                    data = hit.data.to_dict()
                    resource_totals[resource_id] += 1
                    chunk.append(data)
                    if len(chunk) == chunk_size:
                        _flush(chunk)
                        chunk = []
                _flush(chunk)

        non_datastore_resources = [
            k
            for k, v in self.query.resource_ids_and_versions.items()
            if v == common.NON_DATASTORE_VERSION
        ]
        for resource_id in non_datastore_resources:
            resource_totals[resource_id] = 1
            field_counts[resource_id] = {}

        self.core_record.update(
            resource_totals=resource_totals,
            total=sum(filter(None, resource_totals.values())),
            field_counts=field_counts,
        )

        return self.core_record

    def generate_derivative(self):
        """
        Generates derivative files, if necessary.

        :return:
        """

        self.request.update_status(DownloadRequest.state_derivative_gen)

        if self.derivative_record.filepath is None:
            # if this _is_ defined, we don't need to generate the file
            zip_name = f'{self.hash}.zip'
            zip_path = os.path.join(self.download_dir, zip_name)

            self.derivative_record.update(filepath=zip_path)

            # for keeping track of elapsed generation time
            start = dt.utcnow()
            # for storing build files
            temp_dir = tempfile.mkdtemp()
            self._temp.append(temp_dir)

            manifest = {
                'download_id': self.request.id,
                'resources': {},
                'separate_files': self.derivative_options.separate_files,
                'file_format': self.derivative_options.format,
                'format_args': self.derivative_options.format_args,
                'ignore_empty_fields': self.derivative_options.ignore_empty_fields,
                'transform': self.derivative_options.transform,
                'total_records': self.core_record.total,
                'start': start.isoformat(),
                # these get filled in later but we'll initialise them here
                'files': [],
                'end': None,
                'duration_in_seconds': 0,
            }

            if self.derivative_options.format != 'raw':
                # set up the derivative generators; each generator creates one component.
                # components = individual file groups within the main zip, e.g. one CSV for each
                # resource (multiple components), or multiple files comprising a single DarwinCore
                # archive (single component)
                fields = partial(
                    get_fields,
                    self.core_record.field_counts,
                    self.derivative_options.ignore_empty_fields,
                )
                if self.derivative_options.separate_files:
                    components = {
                        rid: get_derivative_generator(
                            self.derivative_options.format,
                            output_dir=temp_dir,
                            fields=fields(resource_id=rid),
                            resource_id=rid,
                            query=self.query,
                            **self.derivative_options.format_args,
                        )
                        for rid in self.query.resource_ids_and_versions
                    }
                else:
                    gen = get_derivative_generator(
                        self.derivative_options.format,
                        output_dir=temp_dir,
                        fields=fields(),
                        query=self.query,
                        **self.derivative_options.format_args,
                    )
                    components = defaultdict(lambda: gen)

                # load transformation functions
                transformations = [
                    get_transformation(t, **targs)
                    for t, targs in (self.derivative_options.transform or {}).items()
                ]

                for (
                    resource_id,
                    version,
                ) in self.query.resource_ids_and_versions.items():
                    # add the resource ID as the message for use in the status page
                    self.request.update_status(
                        DownloadRequest.state_derivative_gen, resource_id
                    )
                    if (
                        len(self.query.resource_ids_and_versions) > 1
                        and self.core_record.resource_totals[resource_id] == 0
                    ):
                        # don't generate empty files unless there's only one resource
                        continue
                    derivative_generator = components[resource_id]
                    core_file_path = os.path.join(
                        self.core_folder_path, f'{resource_id}_{version}.avro'
                    )
                    with derivative_generator, open(core_file_path, 'rb') as core_file:
                        for record in fastavro.reader(core_file):
                            # apply the transformations first
                            for transform in transformations:
                                record = transform(record)
                            # filter out fields that are empty for all records
                            if self.derivative_options.ignore_empty_fields:
                                record = filter_data_fields(
                                    record, self.core_record.field_counts[resource_id]
                                )
                            # then write the record
                            derivative_generator.write(record)

                if self.derivative_options.separate_files:
                    for generator in components.values():
                        generator.cleanup()
                else:
                    components.default_factory().cleanup()

            else:
                resource_show = toolkit.get_action('resource_show')
                for (
                    resource_id,
                    version,
                ) in self.query.resource_ids_and_versions.items():
                    self.request.update_status(
                        DownloadRequest.state_derivative_gen, resource_id
                    )
                    res = resource_show({}, {'id': resource_id})
                    if res.get('url_type') == 'upload':
                        upload = uploader.get_resource_uploader(res)
                        if not upload.storage_path:
                            # if ckan.storage_path is not set
                            raise Exception('Non-datastore uploads are not configured.')
                        filepath = upload.get_path(res['id'])
                        original_filename = res.get('url', '').split('/')[-1]
                        fileext = (
                            os.path.splitext(original_filename)[-1].strip('.')
                            or res['format'].lower()
                        )
                        filename = os.path.extsep.join([resource_id, fileext])
                        shutil.copy2(filepath, os.path.join(temp_dir, filename))
                    else:
                        raise NotImplemented(
                            f'No raw file is available for resource {resource_id}'
                        )

            # generation finished
            self.request.update_status(DownloadRequest.state_packaging)

            end = dt.utcnow()
            manifest['end'] = end.isoformat()

            duration = (end - start).total_seconds()
            manifest['duration_in_seconds'] = duration

            files_to_zip = os.listdir(temp_dir) + ['manifest.json']
            manifest['files'] = files_to_zip

            # allow plugins to make changes
            for plugin in PluginImplementations(IVersionedDatastoreDownloads):
                manifest = plugin.download_modify_manifest(manifest, self.request)

            # write out manifest
            with open(
                os.path.join(temp_dir, 'manifest.json'), 'w', encoding='utf8'
            ) as f:
                json.dump(manifest, f, sort_keys=True, indent=2, ensure_ascii=False)

            # zip everything up
            with zipfile.ZipFile(zip_path, 'w', zipfile.ZIP_DEFLATED, True) as z:
                for filename in files_to_zip:
                    z.write(os.path.join(temp_dir, filename), arcname=filename)

        # auth checks should have already been done, so this should have been removed
        # if not allowed
        if self.server.filename:
            # ensure the custom dir exists
            if not os.path.exists(self.custom_dir):
                os.mkdir(self.custom_dir)
            symlink_path = os.path.join(self.custom_dir, f'{self.server.filename}.zip')
            if (
                os.path.islink(symlink_path)
                and os.path.realpath(symlink_path) != self.derivative_record.filepath
            ):
                # this should only be being used by admins, so we assume it's okay to
                # overwrite old symlinks
                os.remove(symlink_path)
            # test it again and create if necessary
            if not os.path.islink(symlink_path):
                os.symlink(self.derivative_record.filepath, symlink_path)

        return self.derivative_record



from .stats import get_last_ingest


def check_version_is_valid(resource_id, version):
    """
    Checks that the given version is valid for the given resource id. Note that we check
    the ingest version not the indexed version as this is the source of truth about the
    versions of the resource we know about.

    The version must be greater than the latest ingested version or there must not be any ingested
    versions available.

    :param resource_id: the resource's id
    :param version: the version to check
    """
    # retrieve the latest ingested version
    ingest_version = get_last_ingest(resource_id)
    # if there is a current version of the resource data the proposed version must be newer
    return ingest_version is None or version > ingest_version.version



from ckantools.validators import list_of_strings
from ckantools.validators.ivalidators import BaseArgs

from ckan.plugins import toolkit
from ckanext.datastore.logic.schema import json_validator

# grab all the validator functions upfront
boolean_validator = toolkit.get_validator('boolean_validator')
ignore_missing = toolkit.get_validator('ignore_missing')
int_validator = toolkit.get_validator('int_validator')
not_missing = toolkit.get_validator('not_missing')
not_empty = toolkit.get_validator('not_empty')
resource_id_exists = toolkit.get_validator('resource_id_exists')
email_validator = toolkit.get_validator('email_validator')


class QueryArgs(BaseArgs):
    query: dict
    query_version: str
    version: int
    resource_ids: list
    resource_ids_and_versions: dict
    slug_or_doi: str

    fields = {
        'query': [ignore_missing, json_validator],
        'query_version': [ignore_missing, str],
        'version': [ignore_missing, int_validator],
        'resource_ids': [ignore_missing, list_of_strings()],
        'resource_ids_and_versions': [ignore_missing, json_validator],
        'slug_or_doi': [ignore_missing, str],
    }


class DerivativeArgs(BaseArgs):
    format: str
    format_args: dict
    separate_files: bool
    ignore_empty_fields: bool
    transform: dict

    fields = {
        'format': [not_missing, str],
        'format_args': [ignore_missing, json_validator],
        'separate_files': [ignore_missing, boolean_validator],
        'ignore_empty_fields': [ignore_missing, boolean_validator],
        'transform': [ignore_missing, json_validator],
    }

    defaults = {
        'format_args': {},
        'separate_files': False,
        'ignore_empty_fields': False,
        'transform': {},
    }


class ServerArgs(BaseArgs):
    type: str
    type_args: dict
    custom_filename: str

    fields = {
        'type': [ignore_missing, str],
        'type_args': [ignore_missing, json_validator],
        'custom_filename': [ignore_missing, str],
    }

    defaults = {'type': 'direct', 'type_args': {}}


class NotifierArgs(BaseArgs):
    type: str
    type_args: dict

    fields = {
        'type': [ignore_missing, str],
        'type_args': [ignore_missing, json_validator],
    }

    defaults = {'type': 'none', 'type_args': {}}



from .base import BaseFileServer
from ckan.plugins import toolkit
import os


class DirectFileServer(BaseFileServer):
    name = 'direct'

    def serve(self, request):
        site_url = toolkit.config.get('ckan.site_url')

        if self.filename:
            return site_url + f'/downloads/custom/{self.filename}.zip'

        filepath = request.derivative_record.filepath
        filename = os.path.split(filepath)[-1]
        return site_url + f'/downloads/direct/{filename}'



from ckan import model
from ckan.plugins import toolkit, PluginImplementations
from splitgill.indexing.utils import DOC_TYPE

from . import common
from ..interfaces import IVersionedDatastore


def get_latest_version(resource_id):
    """
    Retrieves the latest version of the given resource from the status index.

    :param resource_id: the resource's id
    :return: the version or None if the resource isn't indexed
    """
    index_name = prefix_resource(resource_id)
    return common.SEARCH_HELPER.get_latest_index_versions([index_name]).get(
        index_name, None
    )


def prefix_resource(resource_id):
    """
    Adds the configured prefix to the start of the resource id to get the index name for
    the resource data in elasticsearch.

    :param resource_id: the resource id
    :return: the resource's index name
    """
    return f'{common.CONFIG.elasticsearch_index_prefix}{resource_id}'


def unprefix_index(index_name):
    """
    Removes the configured prefix from the start of the index name to get the resource
    id.

    :param index_name: the index name
    :return: the resource's id
    """
    return index_name[len(common.CONFIG.elasticsearch_index_prefix) :]


def prefix_field(field):
    """
    Prefixes a the given field name with "data.". All data from the resource in
    splitgill is stored under the data key in the elasticsearch record so to avoid end
    users needing to know that all fields should be referenced by their non-
    data.-prefixed name until they are internal to the code and can be prefixed before
    being passed on to splitgill.

    :param field: the field name
    :return: data.<field>
    """
    return f'data.{field}'


def get_public_alias_prefix():
    """
    Returns the prefix to use for the public aliases.

    :return: the public prefix
    """
    return 'pub'


def get_public_alias_name(resource_id):
    """
    Returns the name of the alias which makes gives public access to this resource's
    datastore data. This is just "pub" (retrieved from get_public_alias_prefix above)
    prepended to the normal prefixed index name, for example:

        pubnhm-05ff2255-c38a-40c9-b657-4ccb55ab2feb

    :param resource_id: the resource's id
    :return: the name of the alias
    """
    return f'{get_public_alias_prefix()}{prefix_resource(resource_id)}'


def trim_index_name(index_name):
    """
    Given an index's name, remove the prefix returning the original resource id.

    :param index_name: the name of the index
    :return: the resource id
    """
    return index_name[len(common.CONFIG.elasticsearch_index_prefix) :]


def update_resources_privacy(package):
    """
    Update the privacy of the resources in the datastore associated with the given
    package. If the privacy is already set correctly on each of the resource's indices
    in Elasticsearch this does nothing.

    :param package: the package model object (not the dict!)
    """
    for resource in package.resources:
        update_privacy(resource.id, package.private)


def update_privacy(resource_id, is_private=None):
    """
    Update the privacy of the given resource id in the datastore. If the privacy is
    already set correctly on the resource's index in Elasticsearch this does nothing.

    :param resource_id: the resource's id
    :param is_private: whether the package the resource is in is private or not. This is an optional
                       parameter, if it is left out we look up the resource's package in the
                       database and find out the private setting that way.
    :return: True if modifications were required to update the resource data's privacy, False if not
    """
    if is_private is None:
        resource = model.Resource.get(resource_id)
        is_private = resource.package.private
    if is_private:
        return make_private(resource_id)
    else:
        return make_public(resource_id)


def make_private(resource_id):
    """
    Makes the given resource private in elasticsearch. This is accomplished by removing
    the public alias for the resource. If the resource's base index doesn't exist at
    all, or the alias already doesn't exist, nothing happens.

    :param resource_id: the resource's id
    :return: True if modifications were required to make the resource's data private, False if not
    """
    index_name = prefix_resource(resource_id)
    public_index_name = get_public_alias_name(resource_id)
    if common.ES_CLIENT.indices.exists(index_name):
        if common.ES_CLIENT.indices.exists_alias(index_name, public_index_name):
            common.ES_CLIENT.indices.delete_alias(index_name, public_index_name)
            return True
    return False


def make_public(resource_id):
    """
    Makes the given resource public in elasticsearch. This is accomplished by adding an
    alias to the resource's index. If the resource's base index doesn't exist at all or
    the alias already exists, nothing happens.

    :param resource_id: the resource's id
    :return: True if modifications were required to make the resource's data public, False if not
    """
    index_name = prefix_resource(resource_id)
    public_index_name = get_public_alias_name(resource_id)
    if common.ES_CLIENT.indices.exists(index_name):
        if not common.ES_CLIENT.indices.exists_alias(index_name, public_index_name):
            actions = {
                'actions': [{'add': {'index': index_name, 'alias': public_index_name}}]
            }
            common.ES_CLIENT.indices.update_aliases(actions)
            return True
    return False


def is_resource_read_only(resource_id):
    """
    Loops through the plugin implementations checking if any of them want the given
    resource id to be read only.

    :return: True if the resource should be treated as read only, False if not
    """
    implementations = PluginImplementations(IVersionedDatastore)
    return any(
        plugin.datastore_is_read_only_resource(resource_id)
        for plugin in implementations
    )


class ReadOnlyResourceException(toolkit.ValidationError):
    pass


class InvalidVersionException(toolkit.ValidationError):
    pass


def iter_data_fields(mapping):
    """
    Returns an iterator over the fields defined in the given mapping which yields the
    name of the field and the field's config. The names of the fields are represented by
    tuples allowing nested fields to be represented using their whole path (for example,
    a field at the top level is just ('field', ): {} but a nested one would be ('field',
    'sub'): {}).

    :param mapping: the mapping dict returned from elasticsearch, this should be the first value in
                    the dict after the index name, i.e. the result of get_mapping(index)[index]
    :return: an iterator which yields fields and their configs
    """

    def iter_properties(props, path=None):
        # this is a recursive function which can deal with nested fields
        if path is None:
            path = tuple()
        for field, config in props.items():
            if 'properties' in config:
                for result in iter_properties(
                    config['properties'], path=path + (field,)
                ):
                    yield result
            else:
                yield path + (field,), config

    return iter_properties(
        mapping['mappings'][DOC_TYPE]['properties']['data']['properties']
    )


def get_last_after(hits):
    """
    Retrieves the "sort" value from the last record in passed the list of hits.

    :param hits: a list of hits from an elasticsearch response
    :return: a list or None
    """
    if hits and 'sort' in hits[-1].meta:
        return list(hits[-1].meta['sort'])
    else:
        return None


def is_datastore_resource(resource_id):
    """
    Looks up in elasticsearch whether there is an index for this resource or not and
    returns the boolean result. If there is an index, this is a datastore resource, if
    not it isn't.

    :param resource_id: the resource id
    :return: True if the resource is a datastore resource, False if not
    """
    index_name = prefix_resource(resource_id)
    # check that the index for this resource exists and there is a reference to it in the status
    # index
    return common.ES_CLIENT.indices.exists(
        index_name
    ) and index_name in common.SEARCH_HELPER.get_latest_index_versions([index_name])


def is_datastore_only_resource(resource_url):
    """
    Checks whether the resource url is a datastore only resource url. When uploading
    data directly to the API without using a source file/URL the url of the resource
    will be set to "_datastore_only_resource" to indicate that as such. This function
    checks to see if the resource URL provided is one of these URLs. Note that we check
    a few different scenarios as CKAN has the nasty habit of adding a protocol onto the
    front of these URLs when saving the resource, sometimes.

    :param resource_url: the URL of the resource
    :return: True if the resource is a datastore only resource, False if not
    """
    return (
        resource_url == common.DATASTORE_ONLY_RESOURCE
        or resource_url == f'http://{common.DATASTORE_ONLY_RESOURCE}'
        or resource_url == f'https://{common.DATASTORE_ONLY_RESOURCE}'
    )


def is_ingestible(resource):
    """
    Returns True if the resource can be ingested into the datastore and False if not. To
    be ingestible the resource must either be a datastore only resource (signified by
    the url being set to _datastore_only_resource) or have a format that we can ingest
    (the format field on the resource is used for this, not the URL). If the url is
    None, False is returned. This is technically not possible due to a Resource model
    constraint but it's worth covering off anyway.

    :param resource: the resource dict
    :return: True if it is, False if not
    """
    if resource.get('url', None) is None:
        return False

    resource_format = resource.get('format', None)
    return (
        is_datastore_only_resource(resource['url'])
        or (
            resource_format is not None
            and resource_format.lower() in common.ALL_FORMATS
        )
        or (resource_format is not None and resource_format.lower() == 'zip')
    )



from ckan.plugins import PluginImplementations, toolkit
from ckanext.versioned_datastore.interfaces import IVersionedDatastoreDownloads

from . import derivatives, servers, notifiers, transforms
from functools import partial


def get_derivative_generator(derivative_name, *args, **kwargs):
    gens = {g.name: g for g in derivatives.derivatives}
    for plugin in PluginImplementations(IVersionedDatastoreDownloads):
        gens = plugin.download_derivative_generators(gens)
    derivative_class = gens.get(derivative_name)
    if derivative_class is None:
        raise toolkit.ObjectNotFound(
            f'{derivative_name} is not a registered derivative generator type.'
        )
    derivative = derivative_class(*args, **kwargs)
    return derivative


def get_file_server(server_name, *args, **kwargs):
    srvrs = {s.name: s for s in servers.servers}
    for plugin in PluginImplementations(IVersionedDatastoreDownloads):
        srvrs = plugin.download_file_servers(srvrs)
    server_class = srvrs.get(server_name)
    if server_class is None:
        raise toolkit.ObjectNotFound(
            f'{server_name} is not a registered file server type.'
        )
    server = server_class(*args, **kwargs)
    return server


def get_notifier(notifier_type, *args, **kwargs):
    ntfrs = {n.name: n for n in notifiers.notifiers}
    for plugin in PluginImplementations(IVersionedDatastoreDownloads):
        ntfrs = plugin.download_notifiers(ntfrs)
    notifier_class = ntfrs.get(notifier_type)
    if notifier_class is None:
        raise toolkit.ObjectNotFound(
            f'{notifier_type} is not a registered download notifier type.'
        )
    notifier = notifier_class(*args, **kwargs)
    return notifier


def get_transformation(transform_name, **kwargs):
    trns = {t.name: t for t in transforms.transforms}
    for plugin in PluginImplementations(IVersionedDatastoreDownloads):
        trns = plugin.download_data_transformations(trns)
    transform_class = trns.get(transform_name)
    if transform_class is None:
        raise toolkit.ObjectNotFound(
            f'{transform_name} is not a registered download data transformation function.'
        )
    return transform_class(**kwargs)



from ckantools.decorators import action

from ckan.plugins import toolkit
from .meta import help, schema
from .meta.arg_objects import ServerArgs, NotifierArgs, QueryArgs, DerivativeArgs
from ...lib.downloads.download import DownloadRunManager
from ...model.downloads import DownloadRequest
from ...lib.downloads.notifiers import validate_notifier_args


@action(schema.datastore_queue_download(), help.datastore_queue_download)
def datastore_queue_download(
    context,
    query: QueryArgs,
    file: DerivativeArgs,
    server: ServerArgs = None,
    notifier: NotifierArgs = None,
):
    server = server or ServerArgs(**ServerArgs.defaults)
    notifier = notifier or NotifierArgs(**NotifierArgs.defaults)

    validate_notifier_args(notifier.type, notifier.type_args)

    if server.custom_filename:
        # check if admin
        try:
            toolkit.check_access('datastore_custom_download_filename', context)
        except toolkit.NotAuthorized:
            server.custom_filename = None

    download_runner = DownloadRunManager(
        query_args=query,
        derivative_args=file,
        server_args=server,
        notifier_args=notifier,
    )

    job = toolkit.enqueue_job(
        download_runner.run,
        queue='download',
        title=download_runner.request.created.strftime('%Y-%m-%d %H:%M:%S'),
        rq_kwargs={'timeout': '24h'},
    )

    return {
        'queued_at': job.enqueued_at.isoformat(),
        'job_id': job.id,
        'download_id': download_runner.request.id,
        'status_json': toolkit.url_for(
            'datastore_status.download_status_json',
            download_id=download_runner.request.id,
            qualified=True,
        ),
        'status_html': toolkit.url_for(
            'datastore_status.download_status',
            download_id=download_runner.request.id,
            qualified=True,
        ),
    }


@action(schema.datastore_regenerate_download(), help.datastore_regenerate_download)
def datastore_regenerate_download(
    context,
    download_id,
    server: ServerArgs = None,
    notifier: NotifierArgs = None,
):
    # find the download request first
    original_request = DownloadRequest.get(download_id)
    if original_request is None:
        raise toolkit.ObjectNotFound(f'Download "{download_id}" cannot be found.')

    server_args = ServerArgs.defaults.copy()
    if original_request.server_args is not None:
        server_args.update(**original_request.server_args)
    server = server or ServerArgs(**server_args)

    notifier = notifier or NotifierArgs(**NotifierArgs.defaults)

    query = QueryArgs(
        query=original_request.core_record.query,
        query_version=original_request.core_record.query_version,
        resource_ids_and_versions=original_request.core_record.resource_ids_and_versions,
    )

    file = DerivativeArgs(
        format=original_request.derivative_record.format,
        **original_request.derivative_record.options,
    )

    return datastore_queue_download(context, query, file, server, notifier)



import abc

import codecs
import csv
import numbers
import openpyxl
import xlrd
from cchardet import UniversalDetector
from openpyxl.cell.read_only import EmptyCell

from .exceptions import InvalidId
from .utils import ensure_reset
from ... import common


def get_reader(resource_format):
    """
    Given a format, return a ResourceReader instance for it. If no ResourceReader type
    cannot be matched, None is returned.

    :param resource_format: the format
    :return: a ResourceReader instance or None
    """
    resource_format = resource_format.lower()
    if resource_format in common.CSV_FORMATS:
        return SVReader('excel')
    if resource_format in common.TSV_FORMATS:
        return SVReader('excel-tab')
    if resource_format in common.XLS_FORMATS:
        return XLSReader()
    if resource_format in common.XLSX_FORMATS:
        return XLSXReader()

    return None


class ResourceReader(abc.ABC):
    """
    Abstract class to read fields and rows from a resource.
    """

    def __init__(self, compressible):
        '''
        :param compressible: whether the reader can cope with a gzipped file pointer being passed to
                             the get_fields and get_rows functions.
        '''
        self.compressible = compressible

    @abc.abstractmethod
    def get_fields(self, resource_data_fp):
        """
        Returns a list of field names in the order they were provided in the source
        resource.

        :param resource_data_fp: the file pointer to the source
        :return: a list of field names
        """
        pass

    @abc.abstractmethod
    def _get_rows(self, resource_data_fp):
        """
        Returns a generator which yields rows as dicts.

        :param resource_data_fp: the file pointer to the source
        :return: a generator of row dicts
        """
        pass

    def modify_metadata(self, metadata):
        """
        Allows for modification of the metadata before it is written out in the
        intermediate file. The metadata should be modified in place. By default this
        function does nothing and should be overridden by subclasses.

        :param metadata: the current metadata as a dict
        """
        pass

    def iter_rows(self, resource_data_fp):
        """
        Returns a generator which yields the rows as dicts. Each row is checked for an
        _id field and if one is found its value is converted to an int.

        :param resource_data_fp: the file pointer to the source
        :return: a generator of row dicts
        """
        for row_number, row in enumerate(self._get_rows(resource_data_fp)):
            if '_id' in row:
                try:
                    row['_id'] = int(row['_id'])
                except ValueError as e:
                    raise InvalidId(row_number, row, cause=e)
            yield row


class SVReader(ResourceReader):
    '''
    A *SV reader - handles CSVs and TSVs.
    '''

    def __init__(self, dialect):
        '''
        :param dialect: the dialect of the source, this is passed straight to the csv reader
                        constructor function
        '''
        super(SVReader, self).__init__(True)
        self.dialect = dialect
        self.encoding = None

    def _get_dict_reader(self, resource_data_fp):
        """
        Returns a dict reader for the given source file pointer. If the encoding of the
        source hasn't been guessed yet then this function will read the from the file
        pointer until EOF and guess the encoding. The file pointer is reset to the start
        after this occurs.

        :param resource_data_fp: the file pointer to the source
        :return: a DictReader object
        """
        if self.encoding is None:
            with ensure_reset(resource_data_fp):
                detector = UniversalDetector()
                while True:
                    chunk = resource_data_fp.read(8192)
                    if chunk:
                        detector.feed(chunk)
                    else:
                        detector.close()
                        break

            self.encoding = detector.result['encoding']
            # if the detector failed to work out the encoding (unlikely) or if the encoding it
            # comes up with is ASCII, just default to UTF-8 (UTF-8 is a superset of ASCII)
            if self.encoding is None or self.encoding == 'ASCII':
                self.encoding = 'utf-8'

        # create and return the dict reader
        text_wrapper = codecs.getreader(self.encoding)(resource_data_fp)
        return csv.DictReader(text_wrapper, dialect=self.dialect)

    def _get_rows(self, resource_data_fp):
        """
        Returns a generator which yields rows as dicts using a DictReader. If the
        encoding of the source hasn't been guessed yet then this function will read the
        from the file pointer until EOF and guess the encoding. The file pointer is
        reset to the start after this occurs.

        :param resource_data_fp: the file pointer to the source
        :return: a generator of row dicts
        """
        with ensure_reset(resource_data_fp):
            reader = self._get_dict_reader(resource_data_fp)
            for row in reader:
                yield row

    def get_fields(self, resource_data_fp):
        """
        Returns a list of field names in the order they were provided in the source
        resource. This is achieved by reading the first line from the given file
        pointer.

        :param resource_data_fp: the file pointer to the source
        :return: a list of field names
        """
        with ensure_reset(resource_data_fp):
            reader = self._get_dict_reader(resource_data_fp)
            return reader.fieldnames

    def modify_metadata(self, metadata):
        """
        Modify the metadata to include the original encoding of the file we've read.

        :param metadata: the current metadata as a dict
        """
        metadata['original_encoding'] = self.encoding


class XLSReader(ResourceReader):
    """
    Old style Excel file reader.
    """

    def __init__(self):
        # compressible needs to be false here as xlrd opens the file using it's name, not the byte
        # stream
        super(XLSReader, self).__init__(False)

    def _iter_rows(self, resource_data_fp):
        """
        Returns a generator of rows from the resource file pointer. Each row is a list
        of cells.

        :param resource_data_fp: the file pointer to the source
        :return: a generator of cell lists
        """
        with ensure_reset(resource_data_fp):
            # open the xls file up
            book = xlrd.open_workbook(resource_data_fp.name)
            # select the first sheet by default
            sheet = book.sheet_by_index(0)
            # get_rows is a generator so just return it
            return sheet.get_rows()

    def _get_rows(self, resource_data_fp):
        """
        Returns a generator of row dicts from the source.

        :param resource_data_fp: the file pointer to the source
        :return: a generator of row dicts
        """
        rows = self._iter_rows(resource_data_fp)
        # assume the first row is the header
        header = [str(cell.value) for cell in next(rows)]
        # then read all the other rows as data
        for row in rows:
            data = {}
            for field, cell in zip(header, row):
                # if the cell is the id column, it contains a number and the number is an
                # integer, convert it from a float to an int
                if field == '_id':
                    if cell.ctype == xlrd.XL_CELL_NUMBER and cell.value.is_integer():
                        data[field] = int(cell.value)
                    else:
                        raise Exception('_id not int')
                elif cell == xlrd.XL_CELL_EMPTY:
                    # ignore empty cells
                    continue
                else:
                    # otherwise just use the value (if it's not an empty string)
                    value = str(cell.value)
                    if value:
                        data[field] = str(cell.value)
            # yield the row
            yield data

    def get_fields(self, resource_data_fp):
        """
        Returns a list of field names in the order they were provided in the source
        resource. This is achieved by assuming the first line in the worksheet is the
        header row.

        :param resource_data_fp: the file pointer to the source
        :return: a list of field names
        """
        # assume the first row is the header
        return [str(cell.value) for cell in next(self._iter_rows(resource_data_fp))]


class XLSXReader(ResourceReader):
    """
    New style Excel file reader.
    """

    def __init__(self):
        # compressible needs to be false here as openpyxl can't handle it
        super(XLSXReader, self).__init__(False)

    def _iter_rows(self, resource_data_fp):
        """
        Returns a generator of rows from the resource file pointer. Each row is a list
        of cells.

        :param resource_data_fp: the file pointer to the source
        :return: a generator of cell lists
        """
        with ensure_reset(resource_data_fp):
            workbook = openpyxl.load_workbook(resource_data_fp, read_only=True)
            # get a generator for the rows in the first worksheet and return it
            return workbook.worksheets[0].iter_rows()

    def _get_rows(self, resource_data_fp):
        """
        Returns a generator of row dicts from the source.

        :param resource_data_fp: the file pointer to the source
        :return: a generator of row dicts
        """
        # get a generator for the rows in the first worksheet
        rows = self._iter_rows(resource_data_fp)
        # assume the first row is a header
        header = [str(cell.value) for cell in next(rows)]
        # then read all the other rows as data
        for row in rows:
            data = {}
            for field, cell in zip(header, row):
                # if the cell is the id column and it contains a number make sure it stays a
                # number
                if field == '_id':
                    if isinstance(cell.value, numbers.Integral):
                        data[field] = cell.value
                    else:
                        raise Exception('_id not int')
                # ignore empty cells
                elif isinstance(cell, EmptyCell) or cell.value is None:
                    continue
                else:
                    # convert everything else to unicode
                    value = str(cell.value)
                    if value:
                        data[field] = value
            # yield the row
            yield data

    def get_fields(self, resource_data_fp):
        """
        Returns a list of field names in the order they were provided in the source
        resource. This is achieved by assuming the first line in the worksheet is the
        header row.

        :param resource_data_fp: the file pointer to the source
        :return: a list of field names
        """
        # assume the first row is the header
        return [str(cell.value) for cell in next(self._iter_rows(resource_data_fp))]


class APIReader(ResourceReader):
    """
    Direct API data upload reader.

    The data comes to us as a list of dicts.
    """

    def __init__(self, data):
        '''
        :param data: the data as a list of dicts
        '''
        super(APIReader, self).__init__(False)
        self.data = data

    def _get_rows(self, *args, **kwargs):
        """
        Returns a generator of row dicts from the sent data, this just returns an
        iterator for the data list.

        :return: a generator of row dicts
        """
        return iter(self.data)

    def get_fields(self, *args, **kwargs):
        """
        Returns a list of field names. We can't really guarantee order here because
        python2 doesn't guarantee dict key ordering :(

        :return: a list of field names
        """
        fields = []
        for row in self.data:
            for field in row.keys():
                if field not in fields:
                    fields.append(field)
        return fields



from ckan.plugins import toolkit

from .importing import (
    import_resource_data,
    ResourceImportRequest,
    ResourceDeletionRequest,
    delete_resource_data,
)
from .indexing import index_resource, ResourceIndexRequest


def queue(task, request):
    """
    Generic queueing function which ensures we set common attributes when queueing the
    task.

    :param task: the function to queue
    :param request: the queue request object
    :return: the queued job
    """
    # pass a timeout of 1 hour (3600 seconds)
    return toolkit.enqueue_job(
        task,
        args=[request],
        queue='importing',
        title=str(request),
        rq_kwargs={'timeout': 3600},
    )


def queue_import(resource, version, replace, records=None, api_key=None):
    """
    Queues a job which when run will import the data for the resource.

    :param resource: the resource we're going to import (this must be the resource dict)
    :param version: the version of the resource to import
    :param replace: whether to replace the existing data or not
    :param records: a list of dicts to import, or None if the data is coming from URL or file
    :param api_key: the api key of the user who initiated the import, this is required if the
                    package the resource is in is private and the data in the resource was uploaded
    :return: the queued job
    """
    resource_import_request = ResourceImportRequest(
        resource, version, replace, records, api_key
    )
    return queue(import_resource_data, resource_import_request)


def queue_index(resource, lower_version, upper_version):
    """
    Queues a job which when run will index the data between the given versions for the
    resource.

    :param resource: the resource we're going to index (this must be the resource dict)
    :param lower_version: the lower version to index (exclusive)
    :param upper_version: the upper version to index (inclusive)
    :return: the queued job
    """
    resource_index_request = ResourceIndexRequest(
        resource, lower_version, upper_version
    )
    return queue(index_resource, resource_index_request)


def queue_deletion(resource, version):
    """
    Queues a job which when run will delete all the data in a resource by saving a new
    version into mongo for each record where the data field is empty ({}). After this is
    done, an index is completed.

    :param resource: the resource we're going to delete (this must be the resource dict)
    :param version: the version to delete the data in
    :return: the queued job
    """
    deletion_request = ResourceDeletionRequest(resource, version)
    return queue(delete_resource_data, deletion_request)



import json
import os
import shutil
from uuid import uuid4

import pytest
from mock import patch

from ckanext.versioned_datastore.lib.downloads.derivatives.dwc import urls
from ckanext.versioned_datastore.lib.downloads.derivatives.dwc.schema import Schema

schema_cache = '/tmp/schema_cache'


class TestDwcSchemaSerialisation:
    def setup_method(self):
        os.mkdir(schema_cache)

    def teardown_method(self):
        shutil.rmtree(schema_cache)

    def test_can_serialise_schema(self):
        schema = Schema.regenerate()
        serialised_schema = schema.serialise()
        assert isinstance(serialised_schema, dict)
        assert 'domains' in serialised_schema
        assert 'props' in serialised_schema
        assert (
            serialised_schema['row_type'] == 'http://rs.tdwg.org/dwc/terms/Occurrence'
        )
        assert 'core_extension' not in serialised_schema

    def test_can_save_schema(self):
        schema = Schema.regenerate()
        fn = os.path.extsep.join([uuid4().hex, 'json'])
        fp = os.path.join(schema_cache, fn)
        schema.save(fp)
        assert os.path.exists(fp)

        with open(fp, 'r') as f:
            serialised_schema = json.load(f)

        assert isinstance(serialised_schema, dict)
        assert 'domains' in serialised_schema
        assert 'props' in serialised_schema
        assert (
            serialised_schema['row_type'] == 'http://rs.tdwg.org/dwc/terms/Occurrence'
        )

    def test_can_load_schema(self):
        schema = Schema.regenerate()
        fn = os.path.extsep.join([uuid4().hex, 'json'])
        fp = os.path.join(schema_cache, fn)
        schema.save(fp)
        assert os.path.exists(fp)

        with patch(
            'ckanext.versioned_datastore.lib.downloads.derivatives.dwc.schema.Schema.regenerate'
        ) as regenerate:
            loaded_schema = Schema.load(fp)
            regenerate.assert_not_called()

        assert isinstance(loaded_schema, Schema)
        assert loaded_schema.row_type == 'http://rs.tdwg.org/dwc/terms/Occurrence'

    def test_regenerates_if_not_exists(self):
        fn = os.path.extsep.join([uuid4().hex, 'json'])
        fp = os.path.join(schema_cache, fn)
        assert not os.path.exists(fp)

        with patch(
            'ckanext.versioned_datastore.lib.downloads.derivatives.dwc.schema.Schema.regenerate'
        ) as regenerate:
            loaded_schema = Schema.load(fp)
            regenerate.assert_called()

    @pytest.mark.parametrize('core_extension_url', list(urls.core_extensions.values()))
    def test_serialise_core_extension(self, core_extension_url):
        schema = Schema.regenerate(core_extension_url=core_extension_url)

        assert schema.core_extension.location.url == core_extension_url.url

        serialised_schema = schema.serialise()
        assert (
            serialised_schema['core_extension']['location']['url']
            == core_extension_url.url
        )

    def test_load_serialised_core_extension(self):
        core_extension_url = urls.core_extensions['gbif_occurrence']
        schema = Schema.regenerate(core_extension_url=core_extension_url)
        fn = os.path.extsep.join([uuid4().hex, 'json'])
        fp = os.path.join(schema_cache, fn)
        schema.save(fp)
        assert os.path.exists(fp)

        loaded_schema = Schema.load(fp)

        assert loaded_schema.core_extension.location.url == core_extension_url.url

    @pytest.mark.parametrize('extension_url', list(urls.extensions.values()))
    def test_serialise_extensions(self, extension_url):
        schema = Schema.regenerate(extension_urls=[extension_url])

        assert len(schema.extensions) == 1
        assert schema.extensions[0].location.url == extension_url.url

        serialised_schema = schema.serialise()
        assert len(serialised_schema['extensions']) == 1
        assert (
            serialised_schema['extensions'][0]['location']['url'] == extension_url.url
        )

    def test_load_serialised_extensions(self):
        extension_url = urls.extensions['gbif_multimedia']
        schema = Schema.regenerate(extension_urls=[extension_url])
        fn = os.path.extsep.join([uuid4().hex, 'json'])
        fp = os.path.join(schema_cache, fn)
        schema.save(fp)
        assert os.path.exists(fp)

        loaded_schema = Schema.load(fp)

        assert len(loaded_schema.extensions) == 1
        assert loaded_schema.extensions[0].location.url == extension_url.url



from openpyxl import Workbook, load_workbook

from .base import BaseDerivativeGenerator
from ..utils import flatten_dict


class XlsxDerivativeGenerator(BaseDerivativeGenerator):
    name = 'xlsx'
    extension = 'xlsx'

    DEFAULT_SHEET_NAME = 'Data'

    def __init__(self, output_dir, fields, query, resource_id=None, **format_args):
        super(XlsxDerivativeGenerator, self).__init__(
            output_dir, fields, query, resource_id, **format_args
        )
        self.workbook = None

    def initialise(self):
        self.workbook.create_sheet(self.DEFAULT_SHEET_NAME)
        self.workbook.active.append(self.fields['main'])
        super(XlsxDerivativeGenerator, self).initialise()

    def setup(self):
        try:
            self.workbook = load_workbook(self.file_paths['main'])
        except Exception as e:
            self.workbook = Workbook(write_only=True)
        super(XlsxDerivativeGenerator, self).setup()

    def finalise(self):
        try:
            self.workbook.save(self.file_paths['main'])
        finally:
            # if something goes wrong when trying to save the workbook, make sure to
            # close the workbook before raising the error
            self.workbook.close()
        super(XlsxDerivativeGenerator, self).finalise()

    def _write(self, record):
        row = flatten_dict(record)
        filtered_row = {}
        for field, value in row.items():
            if value is None and field not in self.fields['main']:
                continue
            elif field not in self.fields['main']:
                raise ValueError('Unexpected field.')
            else:
                filtered_row[field] = value
        if self.resource_id:
            filtered_row[self.RESOURCE_ID_FIELD_NAME] = self.resource_id
        self.workbook.active.append(
            filtered_row.get(field) for field in self.fields['main']
        )



import json

from ckan.model import meta, DomainObject
from sqlalchemy import Column, Table, BigInteger, UnicodeText

# this table stores general details about each version of each resource. Currently it only stores
# the column names and order.
datastore_resource_details_table = Table(
    'versioned_datastore_resource_details',
    meta.metadata,
    Column('id', BigInteger, primary_key=True),
    Column('resource_id', UnicodeText, nullable=False, index=True),
    # the version this operation is creating (for ingestion this means the version it's adding to
    # mongo, for indexing this means the version it's pulling from mongo and putting into
    # elasticsearch)
    Column('version', BigInteger, nullable=False),
    # the detailed operation breakdown returned by splitgill
    Column('columns', UnicodeText, nullable=False),
    # a hash of the ingested file - can be None and will be if records are directly added using the
    # API
    Column('file_hash', UnicodeText),
)


class DatastoreResourceDetails(DomainObject):
    """
    Object for holding datastore resource details, currently just the columns at each
    version.
    """

    def get_columns(self, validate=True):
        """
        Retrieve the columns contained in this resource's version.

        :param validate: if True (the default) then fullstops are replaced with underscores before
                         returning the list of columns and any falsey columns (empty strings, Nones)
                         are removed
        :return: a list of column names in the order they were in the original data source
        """
        columns = []
        for column in json.loads(self.columns):
            if validate:
                if not column:
                    continue
                column = column.replace('.', '_')
            columns.append(column)
        return columns


meta.mapper(DatastoreResourceDetails, datastore_resource_details_table)



import json

import pytest
from ckan.plugins import toolkit
from ckanext.versioned_datastore.lib.basic_query.geo import (
    add_point_filter,
    FIELD,
    add_multipolygon_filter,
    add_polygon_filter,
    add_geo_search,
)
from elasticsearch_dsl.query import GeoPolygon, Bool
from mock import MagicMock, call, patch


class TestAddPointFilter(object):
    def test_simple(self):
        search = MagicMock(filter=MagicMock())
        distance = MagicMock()
        coordinates = (4.3, 1.2)

        returned_search = add_point_filter(search, distance, coordinates)

        assert returned_search == search.filter.return_value
        assert search.filter.call_count == 1
        assert search.filter.call_args == call(
            'geo_distance',
            **{
                'distance': distance,
                FIELD: {
                    'lat': coordinates[1],
                    'lon': coordinates[0],
                },
            }
        )

    def test_float_conversion(self):
        search = MagicMock(filter=MagicMock())
        distance = MagicMock()
        coordinates = ('4.53423', 100)

        returned_search = add_point_filter(search, distance, coordinates)

        assert returned_search == search.filter.return_value
        assert search.filter.call_count == 1
        assert search.filter.call_args == call(
            'geo_distance',
            **{
                'distance': distance,
                FIELD: {
                    'lat': float(coordinates[1]),
                    'lon': float(coordinates[0]),
                },
            }
        )


class TestAddMultiPolygonFilter(object):
    def test_simple(self):
        search = MagicMock(filter=MagicMock())
        coordinates = [[[[-16, 44], [-13.1, 34.8], [15.99, 35], [5, 49]]]]

        returned_search = add_multipolygon_filter(search, coordinates)

        assert returned_search == search.filter.return_value
        assert search.filter.call_count == 1

        filters = [
            GeoPolygon(
                **{
                    FIELD: {
                        'points': [
                            {
                                'lat': 44.0,
                                'lon': -16.0,
                            },
                            {
                                'lat': 34.8,
                                'lon': -13.1,
                            },
                            {
                                'lat': 35.0,
                                'lon': 15.99,
                            },
                            {
                                'lat': 49.0,
                                'lon': 5.0,
                            },
                        ]
                    }
                }
            )
        ]
        assert search.filter.call_args == call(
            Bool(should=filters, minimum_should_match=1)
        )

    def test_validation_error(self):
        search = MagicMock(filter=MagicMock())
        # only two points!
        coordinates = [[[[-16, 44], [-13.1, 34.8]]]]

        with pytest.raises(toolkit.ValidationError):
            add_multipolygon_filter(search, coordinates)

    def test_float_conversion(self):
        search = MagicMock(filter=MagicMock())
        coordinates = [
            [[['-16', '44'], ['-13.1', '34.8'], ['15.99', '35'], ['5', '49']]]
        ]

        returned_search = add_multipolygon_filter(search, coordinates)

        assert returned_search == search.filter.return_value
        assert search.filter.call_count == 1

        filters = [
            GeoPolygon(
                **{
                    FIELD: {
                        'points': [
                            {
                                'lat': 44.0,
                                'lon': -16.0,
                            },
                            {
                                'lat': 34.8,
                                'lon': -13.1,
                            },
                            {
                                'lat': 35.0,
                                'lon': 15.99,
                            },
                            {
                                'lat': 49.0,
                                'lon': 5.0,
                            },
                        ]
                    }
                }
            )
        ]
        assert search.filter.call_args == call(
            Bool(should=filters, minimum_should_match=1)
        )


class TestAddPolygonFilter(object):
    def test_pass_off(self):
        # add_polygon_filter just uses add_multipolygon_filter which we already have a test for so
        # we can just test that it is called
        search = MagicMock()
        coordinates = [[['-16', '44'], ['-13.1', '34.8'], ['15.99', '35'], ['5', '49']]]
        mock_add_multipolygon_filter = MagicMock()

        with patch(
            'ckanext.versioned_datastore.lib.basic_query.geo.add_multipolygon_filter',
            mock_add_multipolygon_filter,
        ):
            add_polygon_filter(search, coordinates)

        assert mock_add_multipolygon_filter.call_args == call(search, [coordinates])


class TestAddGeoSearch(object):
    def test_valid_point_dict(self):
        search = MagicMock()
        geo_filter = {
            'type': 'Point',
            'distance': MagicMock(),
            'coordinates': MagicMock(),
        }

        add_mock = MagicMock()
        with patch(
            'ckanext.versioned_datastore.lib.basic_query.geo.add_point_filter', add_mock
        ):
            add_geo_search(search, geo_filter)

        assert add_mock.call_count == 1
        assert add_mock.call_args == call(
            search,
            distance=geo_filter['distance'],
            coordinates=geo_filter['coordinates'],
        )

    def test_valid_point_string(self):
        search = MagicMock()
        geo_filter = {
            'type': 'Point',
            # can't use magic mocks here because they need to be JSON serialised
            'distance': 'distance mock',
            'coordinates': 'coord mock',
        }

        add_mock = MagicMock()
        with patch(
            'ckanext.versioned_datastore.lib.basic_query.geo.add_point_filter', add_mock
        ):
            add_geo_search(search, json.dumps(geo_filter))

        assert add_mock.call_count == 1
        assert add_mock.call_args == call(
            search,
            distance=geo_filter['distance'],
            coordinates=geo_filter['coordinates'],
        )

    def test_invalid_point(self):
        search = MagicMock()

        with pytest.raises(toolkit.ValidationError):
            geo_filter = {
                'type': 'Point',
            }
            add_geo_search(search, geo_filter)

        for param in ('distance', 'coordinates'):
            with pytest.raises(toolkit.ValidationError):
                geo_filter = {
                    'type': 'Point',
                    param: MagicMock(),
                }
                add_geo_search(search, geo_filter)

    def test_valid_multipolygon_dict(self):
        search = MagicMock()
        geo_filter = {
            'type': 'MultiPolygon',
            'coordinates': MagicMock(),
        }

        add_mock = MagicMock()
        with patch(
            'ckanext.versioned_datastore.lib.basic_query.geo.add_multipolygon_filter',
            add_mock,
        ):
            add_geo_search(search, geo_filter)

        assert add_mock.call_count == 1
        assert add_mock.call_args == call(search, coordinates=geo_filter['coordinates'])

    def test_valid_multipolygon_string(self):
        search = MagicMock()
        geo_filter = {
            'type': 'MultiPolygon',
            # can't use magic mocks here because they need to be JSON serialised
            'coordinates': 'coord mock',
        }

        add_mock = MagicMock()
        with patch(
            'ckanext.versioned_datastore.lib.basic_query.geo.add_multipolygon_filter',
            add_mock,
        ):
            add_geo_search(search, json.dumps(geo_filter))

        assert add_mock.call_count == 1
        assert add_mock.call_args == call(search, coordinates=geo_filter['coordinates'])

    def test_invalid_multipolygon(self):
        search = MagicMock()

        with pytest.raises(toolkit.ValidationError):
            geo_filter = {
                'type': 'MultiPolygon',
            }
            add_geo_search(search, geo_filter)

    def test_valid_polygon_dict(self):
        search = MagicMock()
        geo_filter = {
            'type': 'Polygon',
            'coordinates': MagicMock(),
        }

        add_mock = MagicMock()
        with patch(
            'ckanext.versioned_datastore.lib.basic_query.geo.add_polygon_filter',
            add_mock,
        ):
            add_geo_search(search, geo_filter)

        assert add_mock.call_count == 1
        assert add_mock.call_args == call(search, coordinates=geo_filter['coordinates'])

    def test_valid_polygon_string(self):
        search = MagicMock()
        geo_filter = {
            'type': 'Polygon',
            # can't use magic mocks here because they need to be JSON serialised
            'coordinates': 'coord mock',
        }

        add_mock = MagicMock()
        with patch(
            'ckanext.versioned_datastore.lib.basic_query.geo.add_polygon_filter',
            add_mock,
        ):
            add_geo_search(search, json.dumps(geo_filter))

        assert add_mock.call_count == 1
        assert add_mock.call_args == call(search, coordinates=geo_filter['coordinates'])

    def test_invalid_polygon(self):
        search = MagicMock()

        with pytest.raises(toolkit.ValidationError):
            geo_filter = {
                'type': 'Polygon',
            }
            add_geo_search(search, geo_filter)

    def test_invalid_type(self):
        search = MagicMock()

        with pytest.raises(toolkit.ValidationError):
            # a type we don't support
            add_geo_search(search, {'type': 'not a type we support'})

        with pytest.raises(toolkit.ValidationError):
            # no type
            add_geo_search(search, {})

    def test_invalid_json(self):
        search = MagicMock()
        # here's a valid geo_filter value
        geo_filter = {
            'type': 'Polygon',
            # can't use magic mocks here because they need to be JSON serialised
            'coordinates': 'coord mock',
        }
        # but we're gonna add the word beans at the end to make it invalid
        geo_filter = json.dumps(geo_filter) + 'beans'

        with pytest.raises(toolkit.ValidationError):
            add_geo_search(search, geo_filter)



import operator

from ckan.plugins import PluginImplementations
from elasticsearch_dsl import Search

from .geo import add_geo_search
from ..datastore_utils import prefix_field
from ...interfaces import IVersionedDatastore


def _find_version(data_dict):
    """
    Retrieve the version from the data_dict. The version can be specified as a parameter
    in it's own right or as a special filter in the filters dict using the key
    __version__. Using the version parameter is preferred and will override any filter
    version value. The filter method is provided because of limitations in the CKAN
    recline.js framework used by the NHM on CKAN 2.3 where no additional parameters can
    be passed other than q, filters etc.

    :param data_dict: the data dict, this might be modified if the __version__ key is used (it will
                      be removed if present)
    :return: the version found as an integer, or None if no version was found
    """
    version = data_dict.get('version', None)

    # TODO: __version__ support should be removed once the frontend is capable of using the param
    # pop the __version__ to avoid including it in the normal search filters
    filter_version = data_dict.get('filters', {}).pop('__version__', None)
    # it'll probably be a list cause it's a normal filter as far as the frontend is concerned
    if isinstance(filter_version, list):
        # just use the first value
        filter_version = filter_version[0]

    # use the version parameter's value first if it exists
    if version is not None:
        return int(version)
    # otherwise fallback on __version__
    if filter_version is not None:
        return int(filter_version)
    # no version found, return None
    return None


def create_search(context, data_dict, original_data_dict):
    """
    Create the search object based on the parameters in the data_dict. This function
    will call plugins that implement the datastore_modify_data_dict and
    datastore_modify_search interface functions.

    :param context: the context dict
    :param data_dict: the data dict of parameters
    :return: a 3-tuple containing: the original data_dict that was passed into this function, the
                                   data_dict after modification by other plugins and finally the
                                   elasticsearch-dsl Search object
    """
    # allow other extensions implementing our interface to modify the data_dict
    for plugin in PluginImplementations(IVersionedDatastore):
        data_dict = plugin.datastore_modify_data_dict(context, data_dict)

    # extract the version
    version = _find_version(data_dict)

    # create an elasticsearch-dsl Search object by passing the expanded data dict
    search = build_search_object(**data_dict)

    # allow other extensions implementing our interface to modify the search object
    for plugin in PluginImplementations(IVersionedDatastore):
        search = plugin.datastore_modify_search(
            context, original_data_dict, data_dict, search
        )

    return original_data_dict, data_dict, version, search


def build_search_object(
    q=None,
    filters=None,
    after=None,
    offset=None,
    limit=None,
    fields=None,
    facets=None,
    facet_limits=None,
    sort=None,
    **kwargs
):
    """
    Given the parameters, creates a new elasticsearch-dsl Search object and returns it.

    :param q: a query string which will be searched against the meta.all field or a dict of fields
              and search values. If this is a dict then the keys (field names) are always prefixed
              with "data." unless the key is an empty string in which case the field uses is
              meta.all. This allows combination searches across meta.all and data.* fields.
    :param filters: a dict of fields and values to filter the result with. If a key is present that
                    is equal to "__geo__" then the value associated with it should be a dict which
                    will be treated as a geo query to be run against the `meta.geo` field. The value
                    should contain a "type" key which must have a corresponding value of "point",
                    "box" or "polygon" and then other keys that are dependant on the type:
                        - point:
                            - distance: the radius of the circle centred on the specified location
                                        within which records must lie to be matched. This can
                                        specified in any form that elasticsearch accepts for
                                        distances (see their doc, but values like 10km etc).
                            - point: the point to centre the radius on, specified as a lat, long
                                     pair in a list (i.e. [-20, 40.2]).
                        - box:
                            - points: the top left and bottom right points of the box, specified as
                                      a list of two lat/long pairs (i.e. [[-20, 40.2], [0.5, 100]]).
                        - polygon:
                            - points: a list of at least 3 lat/long pairs (i.e. [[-16, 44],
                                      [-13.1, 34.8], [15.99, 35], [5, 49]]).
    :param after: the search after value to start the search result from (for pagination). Cannot be
                  used in conjunction with offset. If both offset and after are provided then after
                  is used and offset is ignored.
    :param offset: the offset to start the search result from (for pagination)
    :param limit: the limit to stop the search result at (for pagination)
    :param fields: a list of field names to return in the result
    :param facets: a list of field names to return an aggregation of top 10 values and counts for
    :param facet_limits: a dict of fields and their customised top n limits
    :param sort: a list of fields to sort by with ordering. By default the fields are sorted
                 ascending, but by providing "desc" after the field name a descending sort will be
                 used. An ascending sort on _id is always added unless included in this list. This
                 is to ensure there is a unique tie-breaking field which is useful for ensuring
                 results stay the same each time they are requested and necessary to ensure the
                 correct result list responses when using the after parameter for pagination.
    :param kwargs: as a convenience we allow a kwargs parameter which we ignore, this is useful to
                   as it allows the arguments to be passed by just unpacking the data_dict
    :return: an elasticsearch-dsl Search object
    """
    search = Search()
    # add a free text query across all fields if there is one. This searches against meta.all which
    # is a copy field created by adding the values of each data.* field
    if q is not None and q != '' and q != {}:
        if isinstance(q, (str, int, float)):
            search = search.query(
                'match', **{'meta.all': {'query': q, 'operator': 'and'}}
            )
        elif isinstance(q, dict):
            for field, query in sorted(q.items(), key=operator.itemgetter(0)):
                # TODO: change this to __all__ to match __geo__?
                if field == '':
                    field = 'meta.all'
                else:
                    field = prefix_field(field)
                search = search.query(
                    'match', **{field: {'query': query, 'operator': 'and'}}
                )
    if filters is not None:
        for field, values in sorted(filters.items(), key=operator.itemgetter(0)):
            if not isinstance(values, list):
                values = [values]
            if field == '__geo__':
                # only pass through the first value
                search = add_geo_search(search, values[0])
            else:
                field = '{}'.format(prefix_field(field))
                for value in values:
                    # filter on the keyword version of the field
                    search = search.filter('term', **{field: value})

    # after and offset cannot be used together, prefer after over offset
    if after is not None:
        search = search.extra(search_after=after)
    elif offset is not None:
        search = search.extra(from_=int(offset))
    # add the limit or a default of 100 if there isn't one specified
    search = search.extra(size=int(limit) if limit is not None else 100)

    if fields is not None:
        search = search.source(list(map(prefix_field, fields)))
    if facets is not None:
        facet_limits = facet_limits if facet_limits is not None else {}
        for facet in facets:
            # to produce the facet counts we use a bucket terms aggregation, note that using the
            # bucket function on the top level aggs attribute on the search object doesn't return a
            # copy of the search object like it does when adding queries etc
            search.aggs.bucket(
                facet,
                'terms',
                field=prefix_field(facet),
                size=facet_limits.get(facet, 10),
            )

    # at least one sort is always added, on the _id column. This is necessary to ensure use of that
    # search_after is predictable (in the elasticsearch docs it recommends that a tie-breaker field
    # is present otherwise the response can include duplicates/missing records). The _id field is
    # always unique and therefore an ideal tie-breaker, so we make sure it's always in the sort
    sorts = []
    # if the caller passes in _id then we don't need to add it in again
    id_in_sort = False
    if sort is not None:
        for field_and_sort in sort:
            if not field_and_sort.endswith(' desc') and not field_and_sort.endswith(
                ' asc'
            ):
                # default the sort direction to ascending if nothing is provided
                field_and_sort += ' asc'
            field, direction = field_and_sort.rsplit(' ', 1)
            # set the id_in_sort boolean to True if we see the _id field in the caller defined sort
            id_in_sort = not id_in_sort and field == '_id'
            field = prefix_field(field)
            # if the sort direction is desc we need to add a minus sign in front of the field name,
            # otherwise we can just use the field name on its own as the default sort is asc
            sorts.append('-{}'.format(field) if direction == 'desc' else field)

    # by default, sort by the _id field
    if not id_in_sort:
        sorts.append(prefix_field('_id'))
    search = search.sort(*sorts)

    return search



import contextlib
import gzip
import logging
import shutil
import tempfile
import zipfile

import codecs
import itertools
import math
import os
import simplejson
import unicodedata
from contextlib import suppress
from datetime import datetime
from splitgill.ingestion.converters import RecordToMongoConverter
from splitgill.ingestion.feeders import IngestionFeeder
from splitgill.ingestion.ingesters import Ingester
from splitgill.mongo import get_mongo

from . import exceptions
from .deletion import ReplaceDeletionFeeder
from .readers import get_reader, APIReader
from .records import DatastoreRecord
from .utils import download_to_temp_file, compute_hash, InclusionTracker
from .. import stats
from ..details import create_details, get_last_file_hash
from ....model.stats import ImportStats

log = logging.getLogger(__name__)


def ingest_resource(version, config, resource, data, replace, api_key):
    """
    Ingest a new version of a resource's data.

    :param version: the new version
    :param config: the splitgill config object
    :param resource: the resource dict
    :param data: the data to ingest (can be None if not using the API)
    :param replace: boolean indicating whether to replace the existing data or not
    :param api_key: the API key if the resource's CKAN URL is to be used as the source and the
                    resource is private
    :return: True if the ingest was successful, False if not
    """
    # cache the resource id as we use it a few times
    resource_id = resource['id']

    log.info(f'Starting validation for {resource_id}')
    # create a stats entry so that preparation progress can be tracked
    prep_stats_id = stats.start_operation(resource_id, stats.PREP, version)
    try:
        data_file_name, data_file_metadata = prepare_resource(
            resource, version, prep_stats_id, data, api_key
        )
    except Exception as e:
        stats.mark_error(prep_stats_id, e)
        log.info(
            f'Prep failed for resource {resource_id} due to {e.__class__.__name__}: {str(e)}'
        )
        if isinstance(e, exceptions.IngestionException):
            # these exceptions are expected (validation problems for example)
            return False
        else:
            raise
    else:
        stats.finish_operation(prep_stats_id, data_file_metadata['record_count'])

    log.info(f'Starting ingest for {resource_id}')
    start = datetime.now()
    # create a stats entry so that progress can be tracked
    stats_id = stats.start_operation(resource_id, stats.INGEST, version, start)
    try:
        feeder = DatastoreFeeder(config, resource_id, version, data_file_name)
        converter = RecordToMongoConverter(version, start)
        ingester = Ingester(version, feeder, converter, config)
        # setup monitoring on the ingester so that we can update the database with stats about the
        # ingestion as it progresses
        stats.monitor_ingestion(stats_id, ingester)

        with InclusionTracker(ingester) as tracker:
            ingester.ingest()

            if replace:
                replace_feeder = ReplaceDeletionFeeder(
                    version, resource_id, tracker, feeder.source
                )
                # note that we use the same converter to ensure the ingestion time is the same
                replace_ingester = Ingester(version, replace_feeder, converter, config)
                # TODO: should we merge the replace ingest stats with the main ingest stats?
                replace_ingester.ingest()
        # we really don't care about errors
        with suppress(Exception):
            # create a details row
            create_details(
                resource_id,
                version,
                data_file_metadata['fields'],
                data_file_metadata['file_hash'],
            )

        return True
    except Exception as e:
        stats.mark_error(stats_id, e)
        log.exception(f'An error occurred during ingestion of {resource_id}')
        return False
    finally:
        # make sure we clean up the intermediate data file
        if os.path.exists(data_file_name):
            os.remove(data_file_name)


def prepare_resource(
    resource, version, stats_id, data=None, api_key=None, update_every=1000
):
    """
    Downloads, validates and then produces an intermediate file containing the rows from
    the resource to be ingested. If the data parameter is provided then it is used as
    the resource data, if not then the resource url is used and the data is extracted
    from the resource url.

    The intermediate file is a JSONL file (newline-delimited JSON using UTF-8 -
    http://jsonlines.org/) which is compressed using GZIP. The first line of the file contains
    metadata about the file including (but not necessary limited to):
        - version
        - resource id
        - fields present (in the order they appear in the source)
        - maximum id present in the source (if any ids were included)
        - record count
        - file hash (will be None if the data parameter is used)

    :param resource: the resource dict
    :param version: the version of the data
    :param stats_id: the stats id for this prep run
    :param data: optional data, if provided must be a list of dicts
    :param api_key: the API key of a user who can read the data, if indeed the data needs an API
                    key to get it. This is needed when the URL is the CKAN resource download URL
                    of a private resource. Can be None to indicate no API key is required
    :param update_every: the frequency with which to update the ImportStats (every x rows written to
                         the intermediate file format). Setting this too low will cause the database
                         to be written to a lot which could cause performance issues
    :return: the name of the intermediate file and the metadata dict
    """
    name = os.path.join(tempfile.gettempdir(), f'{resource["id"]}_{version}.jsonl.gz')

    try:
        with get_fp_and_reader_for_resource_data(resource, data, api_key) as (
            fp,
            reader,
        ):
            # first of all compute a hash of the file so that we can test to see if it's different
            # from the last file we ingested
            last_file_hash = get_last_file_hash(resource['id'])
            if fp is not None:
                file_hash = compute_hash(fp)
            else:
                # rather than hash the data list just use None
                file_hash = None

            # only raise a duplication error if the hashes are the same and they are both non-None
            if file_hash == last_file_hash and file_hash is not None:
                raise exceptions.DuplicateDataSource(file_hash)

            # next, run through the rows in the data to figure out how many rows there are and what
            # the maximum id is (if there is one). If there are validation errors in the data then
            # they will come out here
            record_count = 0
            max_id = -float('inf')
            for row in reader.iter_rows(fp):
                record_count += 1
                if '_id' in row:
                    max_id = max(row['_id'], max_id)

            # this metadata dict will be written out as the first row of the intermediate file
            metadata = {
                'fields': reader.get_fields(fp),
                'file_hash': file_hash,
                'record_count': record_count,
                'resource_id': resource['id'],
                'source': resource['url'] if data is None else 'API',
                'version': version,
            }
            if not math.isinf(max_id):
                # add the max id present if we found one
                metadata['max_id'] = max_id
            # allow the reader to modify the metadata before we write it out
            reader.modify_metadata(metadata)

            # create the intermediate file
            with gzip.open(name, mode='wb') as gzip_file:
                # ensure the data is written out in utf-8
                writer = codecs.getwriter('utf-8')(gzip_file)
                # write the metadata out as a single line first
                writer.write(simplejson.dumps(metadata, ensure_ascii=False) + '\n')
                # then write the rows out as single lines
                for count, row in enumerate(reader.iter_rows(fp), start=1):
                    row_data = simplejson.dumps(row, ensure_ascii=False)
                    # check that the unicode produced doesn't contain any crap characters that we
                    # won't be able to read during ingestion
                    if any(
                        unicodedata.category(character)[0] == 'C'
                        for character in row_data
                    ):
                        raise exceptions.InvalidCharacterException(count, row)

                    writer.write(row_data + '\n')

                    if count % update_every == 0:
                        stats.update_stats(
                            stats_id,
                            {
                                ImportStats.in_progress: True,
                                ImportStats.count: count,
                            },
                        )

            return name, metadata
    except Exception:
        # make sure we clean up the file if there was an error
        if os.path.exists(name):
            os.remove(name)
        raise


@contextlib.contextmanager
def get_fp_and_reader_for_resource_data(resource, data=None, api_key=None):
    """
    Context manager which given a resource, yields a file pointer to the downloaded
    resource source and a ResourceReader instance for reading the data from the file
    pointer.

    If the data parameter is passed (should be a list of dicts) then no file pointer is yielded
    (None is yielded instead). A ResourceReader instance is yielded as normal.

    :param resource: the resource dict
    :param data: optional data, if provided must be a list of dicts
    :param api_key: the API key of a user who can read the data, if indeed the data needs an API
                    key to get it. This is needed when the URL is the CKAN resource download URL
                    of a private resource. Can be None to indicate no API key is required
    :return: yields a file pointer and a ResourceReader instance
    """
    handled = False
    if data is None:
        # there will be a url in the resource dict
        url = resource['url']
        # there may not be a format in the resource dict
        resource_format = resource.get('format', None)
        if resource_format is not None:
            resource_format = resource_format.lower()
        # headers for the download request, note that we only incldue the auth in the request if
        # the url is for an uploaded file, this prevents leaking the credentials
        headers = (
            {'Authorization': api_key}
            if (resource.get('url_type', None) == 'upload' and api_key)
            else {}
        )

        if resource_format != 'zip':
            reader = get_reader(resource_format)
            if reader is not None:
                # if we got a reader, download the resource data to a temporary file
                with download_to_temp_file(
                    url, compress=reader.compressible, headers=headers
                ) as temp:
                    yield temp, reader
                    handled = True
        else:
            # the resource data source is a zip so we need to download the zip, look in it to see if
            # we can find a file to use and then use it. Apologies for the massively nested withs.
            with download_to_temp_file(
                url, compress=False, headers=headers
            ) as temp_download:
                with zipfile.ZipFile(temp_download) as temp_zip:
                    # sort the list of files so that we maintain a consistency when reading zips
                    for name in sorted(temp_zip.namelist()):
                        # use the file extension of the file to see if there's a reader we can use
                        reader = get_reader(os.path.splitext(name)[1][1:])
                        if reader is not None:
                            with temp_zip.open(name, 'r') as zipped_file:
                                with tempfile.NamedTemporaryFile(
                                    mode='w+b', delete=True
                                ) as temp:
                                    # extract the file data from the zip
                                    if reader.compressible:
                                        with gzip.open(temp.name, mode='wb') as g:
                                            shutil.copyfileobj(zipped_file, g)
                                        with gzip.open(temp.name, mode='rb') as g:
                                            yield g, reader
                                    else:
                                        shutil.copyfileobj(zipped_file, temp)
                                        temp.seek(0)
                                        yield temp, reader
                                    handled = True
                                    break
    else:
        yield None, APIReader(data)
        handled = True

    if not handled:
        raise exceptions.UnsupportedDataSource(resource.get('format', None))


class DatastoreFeeder(IngestionFeeder):
    """
    Ingestion feeder class for the versioned datastore.

    This feeder handles reading rows from the intermediate data file format produced by
    the prepare_resource function above.
    """

    def __init__(self, config, resource_id, version, data_file_name):
        '''
        :param config: the splitgill config object
        :param resource_id: the resource id
        :param version: the version of the data we're ingesting
        :param data_file_name: the name of the intermediate data file to read the data from
        '''
        super(DatastoreFeeder, self).__init__(version)
        self.config = config
        self.resource_id = resource_id
        self.data_file_name = data_file_name
        self._header = None

    def iter_rows(self, skip_header_row):
        """
        Returns a generator which can be used to iterate over the rows in the data file.
        This generator yields dicts and therefore handles reading and deserialising the
        rows so that you don't have to.

        :param skip_header_row: whether to skip the header row or yield it as if it was a normal row
        :return: a generator of dicts
        """
        with gzip.open(self.data_file_name, 'rb') as gzip_file:
            # the data file is always in utf-8 format
            reader = codecs.getreader('utf-8')(gzip_file)
            if skip_header_row:
                # skip the first line
                next(reader)
            for line in reader:
                yield simplejson.loads(line)

    def get_existing_max_id(self):
        """
        Figure out what the current max id is in this resource's collection.

        :return: the highest id in the collection currently (it'll be an int), or 0 if there
                aren't any documents in the collection
        """
        with get_mongo(self.config, collection=self.resource_id) as mongo:
            # sort by id descending to get the highest
            doc_with_max_id = mongo.find_one(sort=[('id', -1)])
            # find_one returns None if there aren't any matching documents
            if doc_with_max_id is None:
                return 0
            else:
                return doc_with_max_id['id']

    @property
    def header(self):
        """
        Property providing access to the header in the data file. This function caches
        the header dict in self._header for repeated use and thus avoids reading the
        file every time.

        :return: the header dict
        """
        if self._header is None:
            # load the header by just iterating over the first row in the data file
            self._header = next(self.iter_rows(skip_header_row=False))
        return self._header

    @property
    def source(self):
        """
        Returns the source of the ingestion. This comes from the data file.

        :return: the source name
        """
        return self.header['source']

    def records(self):
        """
        Returns a generator of DatastoreRecords using the rows in the data file.

        The id of each yielded record is determined using either the _id field in the row or by an
        incrementing counter. The counter starts at the maximum id we're aware of - either the
        highest id in the existing mongo collection (if there is one) or the highest id from the _id
        fields in the rows (if there is such a field). Using the highest value of these two sources
        allows us to handle data files where the case where the _id field is only included in some
        rows in the data file. For example, with this data:

            _id, field1, ...
            1,   some data
            2,   some other data
            ,    this one has no id
            3,   but this one does

        if we didn't account for the highest id in the data file the row without an _id (row 3)
        would be assigned _id = 3, but this would then be used by the next row which specifically
        says "I'm row 3" and thus clashing ids!

        :return: a generator of DatastoreRecords
        """
        # calculate the highest id we know about
        highest_id = max(self.get_existing_max_id(), self.header.get('max_id', 0))
        # create a generator which starts at the next number after the highest id
        new_id_generator = itertools.count(start=highest_id + 1, step=1)

        for data in self.iter_rows(skip_header_row=True):
            # see if there's an id defined by the row
            record_id = data.pop('_id', None)
            if record_id is None:
                # create a new id for this record
                record_id = next(new_id_generator)

            yield DatastoreRecord(self.version, record_id, data, self.resource_id)



import logging

from splitgill.config import Config
from splitgill.search import SearchHelper

log = logging.getLogger(__name__)
# stop elasticsearch from showing warning logs
logging.getLogger('elasticsearch').setLevel(logging.ERROR)


# if the resource has been side loaded into the datastore then this should be its URL
DATASTORE_ONLY_RESOURCE = '_datastore_only_resource'
# the formats we support for ingestion
CSV_FORMATS = ['csv', 'application/csv']
TSV_FORMATS = ['tsv']
XLS_FORMATS = ['xls', 'application/vnd.ms-excel']
XLSX_FORMATS = [
    'xlsx',
    'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet',
]
ALL_FORMATS = CSV_FORMATS + TSV_FORMATS + XLS_FORMATS + XLSX_FORMATS

# global variables to hold the splitgill config (not the CKAN one), the splitgill search helper object and
# an elasticsearch client object
CONFIG = None
SEARCH_HELPER = None
ES_CLIENT = None

NON_DATASTORE_VERSION = -1


def setup(ckan_config):
    """
    Given the CKAN config, setup the plugin's global variables.

    :param ckan_config: the ckan config
    """
    global CONFIG, SEARCH_HELPER, ES_CLIENT

    es_hosts = ckan_config.get('ckanext.versioned_datastore.elasticsearch_hosts').split(
        ','
    )
    es_port = ckan_config.get('ckanext.versioned_datastore.elasticsearch_port')
    prefix = ckan_config.get('ckanext.versioned_datastore.elasticsearch_index_prefix')
    CONFIG = Config(
        elasticsearch_hosts=[f'http://{host}:{es_port}/' for host in es_hosts],
        elasticsearch_index_prefix=prefix,
        mongo_host=ckan_config.get('ckanext.versioned_datastore.mongo_host'),
        mongo_port=int(ckan_config.get('ckanext.versioned_datastore.mongo_port')),
        mongo_database=ckan_config.get('ckanext.versioned_datastore.mongo_database'),
    )
    SEARCH_HELPER = SearchHelper(CONFIG)
    # for convenience, expose the client in the search helper at the module level
    ES_CLIENT = SEARCH_HELPER.client



import requests

from ckan.plugins import toolkit
from .base import BaseNotifier


class WebhookNotifier(BaseNotifier):
    name = 'webhook'

    def __init__(
        self, request, url, url_param='url', text_param='text', post=False, **kwargs
    ):
        self.url = url
        self.url_param = url_param
        self.text_param = text_param
        self.post = post
        super(WebhookNotifier, self).__init__(request, **kwargs)

    def notify_start(self):
        request_method = 'POST' if self.post else 'GET'
        text, _ = self.start_text()
        context = self.template_context()
        params = {self.url_param: context['status_page'], self.text_param: text}
        requests.request(request_method, self.url, params=params)

    def notify_end(self, file_url):
        request_method = 'POST' if self.post else 'GET'
        text, _ = self.end_text(file_url)
        params = {self.url_param: file_url, self.text_param: text}
        requests.request(request_method, self.url, params=params)

    def notify_error(self):
        request_method = 'POST' if self.post else 'GET'
        text, _ = self.error_text()
        context = self.template_context()
        params = {self.url_param: context['status_page'], self.text_param: text}
        requests.request(request_method, self.url, params=params)

    @classmethod
    def validate_args(cls, type_args):
        url = type_args.get('url')
        if not url:
            raise toolkit.Invalid('URL must be provided')
        return True



import csv

from .base import BaseDerivativeGenerator
import json


class JsonDerivativeGenerator(BaseDerivativeGenerator):
    name = 'json'
    extension = 'json'

    def __init__(self, output_dir, fields, query, resource_id=None, **format_args):
        super(JsonDerivativeGenerator, self).__init__(
            output_dir, fields, query, resource_id, **format_args
        )
        self._first_row = True

    def initialise(self):
        self.main_file.write('[')
        self._first_row = True
        super(JsonDerivativeGenerator, self).initialise()

    def finalise(self):
        self.main_file.write('\n]')
        super(JsonDerivativeGenerator, self).finalise()

    def _write(self, record):
        if not self._first_row:
            self.main_file.write(',\n')
        else:
            self.main_file.write('\n')
            self._first_row = False

        if self.resource_id:
            record[self.RESOURCE_ID_FIELD_NAME] = self.resource_id

        json_text = json.dumps(record, indent=2)
        indented_json_text = '\n'.join([f'  {line}' for line in json_text.split('\n')])
        self.main_file.write(indented_json_text)



from ckan.plugins import toolkit
from ckantools.decorators import auth


@auth('resource_show', {'resource_id': 'id'}, anon=True)
def datastore_search(context, data_dict):
    return {'success': True}


@auth('resource_update', {'resource_id': 'id'})
def datastore_create(context, data_dict):
    return {'success': True}


@auth('resource_update', {'resource_id': 'id'})
def datastore_upsert(context, data_dict):
    return {'success': True}


@auth('resource_update', {'resource_id': 'id'})
def datastore_delete(context, data_dict):
    return {'success': True}


@auth('resource_show', {'resource_id': 'id'}, anon=True)
def datastore_get_record_versions(context, data_dict):
    return {'success': True}


@auth('resource_show', {'resource_id': 'id'}, anon=True)
def datastore_get_resource_versions(context, data_dict):
    return {'success': True}


@auth('resource_show', {'resource_id': 'id'}, anon=True)
def datastore_autocomplete(context, data_dict):
    return {'success': True}


@auth('resource_update', {'resource_id': 'id'})
def datastore_reindex(context, data_dict):
    return {'success': True}


@auth('resource_show', {'resource_id': 'id'}, anon=True)
def datastore_query_extent(context, data_dict):
    return {'success': True}


@auth('resource_show', {'resource_id': 'id'}, anon=True)
def datastore_get_rounded_version(context, data_dict):
    return {'success': True}


@auth('resource_show', {'resource_id': 'id'}, anon=True)
def datastore_search_raw(context, data_dict):
    return {'success': True}


@auth(anon=True)
def datastore_ensure_privacy(context, data_dict):
    # allow access to everyone
    return {'success': True}


@auth(anon=True)
def datastore_multisearch(context, data_dict):
    # allow access to everyone
    return {'success': True}


@auth(anon=True)
def datastore_field_autocomplete(context, data_dict):
    # allow access to everyone
    return {'success': True}


@auth(anon=True)
def datastore_value_autocomplete(context, data_dict):
    # allow access to everyone
    return {'success': True}


@auth(anon=True)
def datastore_create_slug(context, data_dict):
    # allow access to everyone
    return {'success': True}


@auth(anon=True)
def datastore_resolve_slug(context, data_dict):
    # allow access to everyone
    return {'success': True}


@auth(anon=True)
def datastore_queue_download(context, data_dict):
    # allow access to everyone
    return {'success': True}


@auth(anon=True)
def datastore_regenerate_download(context, data_dict):
    # allow access to everyone
    return {'success': True}


@auth(anon=True)
def datastore_guess_fields(context, data_dict):
    # allow access to everyone
    return {'success': True}


@auth(anon=True)
def datastore_hash_query(context, data_dict):
    # allow access to everyone
    return {'success': True}


@auth(anon=True)
def datastore_is_datastore_resource(context, data_dict):
    # allow access to everyone
    return {'success': True}


@auth(anon=True)
def datastore_get_latest_query_schema_version(context, data_dict):
    # allow access to everyone
    return {'success': True}


@auth(anon=True)
def datastore_count(context, data_dict):
    # allow access to everyone
    return {'success': True}


@auth()
def datastore_edit_slug(context, data_dict):
    # only allows logged-in users
    return {'success': True}


@auth()
def datastore_custom_download_filename(context, data_dict):
    # only allow access to admins (they usually skip this check)
    user_is_sysadmin = context.get('auth_user_obj').sysadmin
    return {'success': user_is_sysadmin}


@auth(anon=True)
def datastore_multisearch_counts(context, data_dict):
    # allow access to everyone
    return {"success": True}



adjectives = [
    'acceptable',
    'according',
    'accurate',
    'active',
    'actual',
    'additional',
    'administrative',
    'adult',
    'afraid',
    'afternoon',
    'aggressive',
    'alone',
    'alternative',
    'amazing',
    'angry',
    'annual',
    'anxious',
    'any',
    'appropriate',
    'asleep',
    'automatic',
    'available',
    'aware',
    'background',
    'basic',
    'beautiful',
    'best',
    'big',
    'bitter',
    'boring',
    'brave',
    'brief',
    'bright',
    'brilliant',
    'broad',
    'budget',
    'business',
    'busy',
    'calm',
    'capable',
    'careful',
    'certain',
    'chance',
    'cheap',
    'chemical',
    'civil',
    'classic',
    'clean',
    'clear',
    'cold',
    'comfortable',
    'commercial',
    'common',
    'competitive',
    'complete',
    'complex',
    'comprehensive',
    'confident',
    'conscious',
    'consistent',
    'constant',
    'content',
    'cool',
    'correct',
    'creative',
    'critical',
    'cultural',
    'curious',
    'current',
    'cute',
    'dangerous',
    'dark',
    'dear',
    'decent',
    'deep',
    'dependent',
    'designer',
    'desperate',
    'different',
    'difficult',
    'direct',
    'dirty',
    'distinct',
    'double',
    'downtown',
    'dramatic',
    'drunk',
    'dry',
    'easy',
    'educational',
    'effective',
    'efficient',
    'electrical',
    'electronic',
    'embarrassed',
    'emergency',
    'emotional',
    'empty',
    'enough',
    'entire',
    'environmental',
    'equal',
    'equivalent',
    'every',
    'exact',
    'excellent',
    'exciting',
    'expensive',
    'expert',
    'express',
    'external',
    'extra',
    'extreme',
    'fair',
    'false',
    'familiar',
    'famous',
    'fast',
    'fat',
    'federal',
    'final',
    'financial',
    'fine',
    'firm',
    'first',
    'fit',
    'flat',
    'formal',
    'former',
    'forward',
    'free',
    'frequent',
    'fresh',
    'friendly',
    'full',
    'fun',
    'funny',
    'future',
    'general',
    'glad',
    'glass',
    'global',
    'gold',
    'good',
    'grand',
    'great',
    'green',
    'gross',
    'guilty',
    'happy',
    'healthy',
    'heavy',
    'helpful',
    'high',
    'historical',
    'honest',
    'huge',
    'hungry',
    'ideal',
    'ill',
    'illegal',
    'immediate',
    'important',
    'impossible',
    'impressive',
    'independent',
    'individual',
    'inevitable',
    'informal',
    'initial',
    'inner',
    'intelligent',
    'interesting',
    'internal',
    'international',
    'junior',
    'key',
    'kind',
    'kitchen',
    'known',
    'large',
    'last',
    'late',
    'latter',
    'leading',
    'least',
    'left',
    'legal',
    'little',
    'local',
    'logical',
    'lonely',
    'long',
    'lost',
    'loud',
    'low',
    'lower',
    'lucky',
    'main',
    'major',
    'many',
    'massive',
    'material',
    'maximum',
    'mean',
    'medical',
    'medium',
    'middle',
    'minor',
    'minute',
    'mobile',
    'mother',
    'narrow',
    'nasty',
    'national',
    'natural',
    'nearby',
    'neat',
    'necessary',
    'negative',
    'neither',
    'nervous',
    'new',
    'nice',
    'normal',
    'novel',
    'numerous',
    'objective',
    'obvious',
    'odd',
    'official',
    'ok',
    'old',
    'open',
    'opening',
    'opposite',
    'orange',
    'ordinary',
    'original',
    'other',
    'overall',
    'parking',
    'particular',
    'party',
    'past',
    'patient',
    'perfect',
    'personal',
    'physical',
    'plastic',
    'pleasant',
    'political',
    'poor',
    'popular',
    'positive',
    'possible',
    'potential',
    'powerful',
    'practical',
    'pregnant',
    'present',
    'pretend',
    'pretty',
    'previous',
    'primary',
    'prior',
    'private',
    'prize',
    'professional',
    'proper',
    'proud',
    'psychological',
    'public',
    'pure',
    'purple',
    'quick',
    'quiet',
    'rare',
    'raw',
    'ready',
    'real',
    'realistic',
    'reasonable',
    'recent',
    'regular',
    'relevant',
    'remarkable',
    'remote',
    'representative',
    'resident',
    'responsible',
    'rich',
    'rough',
    'round',
    'routine',
    'royal',
    'sad',
    'safe',
    'salty',
    'scared',
    'sea',
    'secret',
    'secure',
    'select',
    'senior',
    'sensitive',
    'serious',
    'several',
    'severe',
    'sharp',
    'short',
    'sick',
    'significant',
    'silly',
    'silver',
    'similar',
    'simple',
    'single',
    'slight',
    'slow',
    'small',
    'smart',
    'smooth',
    'soft',
    'solid',
    'sorry',
    'southern',
    'spare',
    'special',
    'specialist',
    'specific',
    'spiritual',
    'square',
    'standard',
    'stock',
    'strange',
    'strict',
    'strong',
    'substantial',
    'successful',
    'such',
    'sudden',
    'sufficient',
    'suitable',
    'super',
    'suspicious',
    'sweet',
    'swimming',
    'tall',
    'technical',
    'temporary',
    'terrible',
    'thin',
    'tiny',
    'top',
    'total',
    'tough',
    'traditional',
    'training',
    'trick',
    'true',
    'typical',
    'ugly',
    'unable',
    'unfair',
    'unhappy',
    'unique',
    'united',
    'unlikely',
    'unusual',
    'upper',
    'upset',
    'used',
    'useful',
    'usual',
    'valuable',
    'various',
    'vast',
    'visible',
    'visual',
    'warm',
    'waste',
    'weak',
    'weekly',
    'weird',
    'whole',
    'wide',
    'wild',
    'willing',
    'winter',
    'wise',
    'wonderful',
    'wooden',
    'work',
    'working',
    'worth',
    'wrong',
    'young',
]

animals = [
    'aardvark',
    'albatross',
    'alligator',
    'angelfish',
    'ant',
    'anteater',
    'antelope',
    'armadillo',
    'axolotl',
    'baboon',
    'badger',
    'bandicoot',
    'barnacle',
    'bat',
    'bear',
    'beaver',
    'beetle',
    'bird',
    'bison',
    'bobcat',
    'bonobo',
    'budgerigar',
    'buffalo',
    'butterfly',
    'camel',
    'capybara',
    'cassowary',
    'cat',
    'caterpillar',
    'catfish',
    'centipede',
    'chameleon',
    'cheetah',
    'chicken',
    'chimpanzee',
    'chinchilla',
    'chipmunk',
    'coral',
    'cougar',
    'cow',
    'coyote',
    'crab',
    'crane',
    'crocodile',
    'cuttlefish',
    'deer',
    'dingo',
    'dodo',
    'dog',
    'dolphin',
    'donkey',
    'dormouse',
    'dragonfly',
    'duck',
    'dugong',
    'eagle',
    'earwig',
    'elephant',
    'emu',
    'falcon',
    'ferret',
    'fish',
    'flamingo',
    'fly',
    'fox',
    'frog',
    'gecko',
    'gerbil',
    'gibbon',
    'giraffe',
    'goat',
    'goose',
    'gorilla',
    'grasshopper',
    'guppy',
    'hamster',
    'hare',
    'hedgehog',
    'heron',
    'hippopotamus',
    'horse',
    'hummingbird',
    'hyena',
    'iguana',
    'impala',
    'jackal',
    'jaguar',
    'jellyfish',
    'kangaroo',
    'kingfisher',
    'kiwi',
    'koala',
    'ladybird',
    'lemming',
    'lemur',
    'leopard',
    'lion',
    'lizard',
    'llama',
    'lobster',
    'lynx',
    'macaw',
    'magpie',
    'manatee',
    'mayfly',
    'meerkat',
    'millipede',
    'mole',
    'mongoose',
    'monkey',
    'moorhen',
    'moose',
    'moth',
    'mouse',
    'mule',
    'newt',
    'nightingale',
    'ocelot',
    'octopus',
    'orangutan',
    'ostrich',
    'otter',
    'oyster',
    'pangolin',
    'panther',
    'parrot',
    'peacock',
    'pelican',
    'penguin',
    'pheasant',
    'pig',
    'piranha',
    'platypus',
    'porcupine',
    'possum',
    'prawn',
    'puffin',
    'puma',
    'quail',
    'rabbit',
    'raccoon',
    'rat',
    'rattlesnake',
    'reindeer',
    'rhinoceros',
    'robin',
    'salamander',
    'scorpion',
    'seahorse',
    'seal',
    'sheep',
    'shrimp',
    'skunk',
    'sloth',
    'snail',
    'snake',
    'sparrow',
    'sponge',
    'squid',
    'squirrel',
    'starfish',
    'stingray',
    'swan',
    'tapir',
    'termite',
    'tiger',
    'tortoise',
    'toucan',
    'turkey',
    'vulture',
    'wallaby',
    'walrus',
    'warthog',
    'wasp',
    'weasel',
    'wildebeest',
    'wolf',
    'wolverine',
    'wombat',
    'woodlouse',
    'woodpecker',
    'yak',
    'zebra',
]

adverbs = [
    'absentmindedly',
    'acceptably',
    'accidentally',
    'accordingly',
    'accurately',
    'actively',
    'actually',
    'additionally',
    'administratively',
    'aggressively',
    'alternatively',
    'always',
    'amazingly',
    'angrily',
    'annually',
    'anxiously',
    'appropriately',
    'automatically',
    'awkwardly',
    'badly',
    'basically',
    'beautifully',
    'bitterly',
    'blindly',
    'boastfully',
    'boldly',
    'bravely',
    'briefly',
    'brightly',
    'brilliantly',
    'broadly',
    'busily',
    'calmly',
    'capably',
    'carefully',
    'certainly',
    'cheaply',
    'cheerfully',
    'chemically',
    'classically',
    'cleanly',
    'clearly',
    'coldly',
    'comfortably',
    'commercially',
    'commonly',
    'competitively',
    'completely',
    'comprehensively',
    'confidently',
    'consciously',
    'consistently',
    'constantly',
    'continually',
    'coolly',
    'correctly',
    'coyly',
    'creatively',
    'critically',
    'culturally',
    'curiously',
    'currently',
    'dangerously',
    'darkly',
    'dearly',
    'decently',
    'deeply',
    'defiantly',
    'deftly',
    'deliberately',
    'delightfully',
    'desperately',
    'devotedly',
    'differently',
    'difficultly',
    'directly',
    'dirtily',
    'distinctly',
    'doubly',
    'doubtfully',
    'dramatically',
    'dryly',
    'dutifully',
    'eagerly',
    'easily',
    'educationally',
    'effectively',
    'efficiently',
    'electrically',
    'electronically',
    'elegantly',
    'emotionally',
    'enormously',
    'entirely',
    'environmentally',
    'equally',
    'essentially',
    'evenly',
    'eventually',
    'exactly',
    'excellently',
    'excitingly',
    'expensively',
    'expertly',
    'expressly',
    'externally',
    'extremely',
    'fairly',
    'faithfully',
    'falsely',
    'famously',
    'federally',
    'ferociously',
    'fiercely',
    'finally',
    'financially',
    'finely',
    'firmly',
    'firstly',
    'flatly',
    'foolishly',
    'formally',
    'formerly',
    'fortunately',
    'freely',
    'frequently',
    'freshly',
    'friendlily',
    'fully',
    'funnily',
    'generally',
    'generously',
    'gladly',
    'gleefully',
    'globally',
    'goodly',
    'gracefully',
    'grandly',
    'greatly',
    'grossly',
    'happily',
    'hastily',
    'healthily',
    'heavily',
    'helpfully',
    'highly',
    'historically',
    'honestly',
    'hopefully',
    'hopelessly',
    'hugely',
    'hungrily',
    'ideally',
    'illegally',
    'immediately',
    'importantly',
    'impossibly',
    'impressively',
    'independently',
    'individually',
    'inevitably',
    'informally',
    'initially',
    'innerly',
    'innocently',
    'inquisitively',
    'intelligently',
    'intensely',
    'interestingly',
    'internally',
    'internationally',
    'irritably',
    'jealously',
    'joyfully',
    'justly',
    'kindly',
    'knowingly',
    'largely',
    'lastly',
    'lately',
    'latterly',
    'lazily',
    'legally',
    'locally',
    'logically',
    'longingly',
    'loosely',
    'loudly',
    'lowly',
    'loyally',
    'luckily',
    'mainly',
    'majorly',
    'massively',
    'materially',
    'medically',
    'merrily',
    'minutely',
    'mockingly',
    'mortally',
    'motherly',
    'mysteriously',
    'narrowly',
    'nastily',
    'nationally',
    'naturally',
    'neatly',
    'necessarily',
    'negatively',
    'nervously',
    'never',
    'newly',
    'nicely',
    'noisily',
    'normally',
    'obediently',
    'objectively',
    'obnoxiously',
    'obviously',
    'occasionally',
    'oddly',
    'offensively',
    'officially',
    'often',
    'only',
    'openly',
    'ordinarily',
    'originally',
    'partially',
    'particularly',
    'patiently',
    'perfectly',
    'personally',
    'physically',
    'plastically',
    'pleasantly',
    'politely',
    'politically',
    'poorly',
    'popularly',
    'positively',
    'possibly',
    'potentially',
    'powerfully',
    'practically',
    'presently',
    'prettily',
    'previously',
    'primarily',
    'privately',
    'professionally',
    'promptly',
    'properly',
    'proudly',
    'psychologically',
    'publicly',
    'punctually',
    'purely',
    'purply',
    'quickly',
    'quietly',
    'quizzically',
    'rapidly',
    'rarely',
    'readily',
    'realistically',
    'really',
    'reasonably',
    'recently',
    'recklessly',
    'regrettably',
    'regularly',
    'remarkably',
    'remotely',
    'repeatedly',
    'responsibly',
    'richly',
    'roughly',
    'roundly',
    'routinely',
    'royally',
    'rudely',
    'sadly',
    'safely',
    'saltily',
    'scarcely',
    'scarily',
    'secretly',
    'securely',
    'seldom',
    'selfishly',
    'sensitively',
    'seriously',
    'severally',
    'severely',
    'shakily',
    'sharply',
    'shortly',
    'sickly',
    'significantly',
    'silently',
    'similarly',
    'simply',
    'singly',
    'slightly',
    'slowly',
    'smartly',
    'smoothly',
    'softly',
    'solemnly',
    'solidly',
    'sometimes',
    'specially',
    'specifically',
    'speedily',
    'spiritually',
    'squarely',
    'sternly',
    'strangely',
    'strictly',
    'strongly',
    'substantially',
    'successfully',
    'suddenly',
    'sufficiently',
    'suitably',
    'suspiciously',
    'sweetly',
    'swimmingly',
    'sympathetically',
    'tally',
    'technically',
    'tediously',
    'temporarily',
    'terribly',
    'thankfully',
    'thinly',
    'totally',
    'traditionally',
    'tremendously',
    'triumphantly',
    'typically',
    'unbearably',
    'unexpectedly',
    'unfairly',
    'unhappily',
    'uniquely',
    'unusually',
    'usefully',
    'usually',
    'vaguely',
    'valiantly',
    'variously',
    'vastly',
    'victoriously',
    'violently',
    'virtually',
    'visibly',
    'visually',
    'vivaciously',
    'warmly',
    'weakly',
    'wearily',
    'weirdly',
    'wholly',
    'widely',
    'wildly',
    'willingly',
    'wisely',
    'woefully',
    'wonderfully',
    'woodenly',
    'wrongly',
]

plants = [
    'almond',
    'apricot',
    'bamboo',
    'banana',
    'baobab',
    'bean',
    'beech',
    'birch',
    'bluebell',
    'cabbage',
    'carrot',
    'cedar',
    'cherry',
    'chestnut',
    'clover',
    'coconut',
    'coffee',
    'elderberry',
    'eucalyptus',
    'fennel',
    'fern',
    'fig',
    'flax',
    'foxglove',
    'fuchsia',
    'garlic',
    'grapefruit',
    'grass',
    'hydrangea',
    'juniper',
    'knotweed',
    'lavender',
    'leek',
    'lemon',
    'lettuce',
    'lupin',
    'magnolia',
    'mahogany',
    'maize',
    'mango',
    'maple',
    'mesquite',
    'milkweed',
    'mulberry',
    'nettle',
    'nightshade',
    'oak',
    'onion',
    'orange',
    'orchid',
    'parsley',
    'parsnip',
    'pea',
    'peach',
    'peanut',
    'pine',
    'pineapple',
    'pistachio',
    'plantain',
    'poplar',
    'potato',
    'primrose',
    'ragweed',
    'raspberry',
    'rhubarb',
    'rice',
    'rocket',
    'rye',
    'saffron',
    'snowdrop',
    'strawberry',
    'sugarcane',
    'sunflower',
    'sugarplum',
    'sycamore',
    'tea',
    'thistle',
    'thyme',
    'tomato',
    'tulip',
    'violet',
    'walnut',
    'wheat',
    'yarrow',
    'yam',
]

list_one = adjectives + adverbs
list_two = adjectives
list_three = animals + plants



import contextlib

import pytest
from elasticsearch_dsl import Search
from mock import MagicMock, patch

from ckanext.versioned_datastore.lib.basic_query.search import (
    _find_version,
    create_search,
    build_search_object,
)
from ckanext.versioned_datastore.lib.datastore_utils import prefix_field


class TestFindVersion(object):
    plugins = ['versioned_datastore']

    def test_none_found(self):
        assert _find_version({}) == None

    def test_version_is_none(self):
        assert _find_version({'version': None}) == None

    def test_version_exists(self):
        assert _find_version({'version': 10}) == 10
        assert _find_version({'version': '10'}) == 10

        with pytest.raises(ValueError):
            _find_version({'version': 'aaaa'})

    def test_version_as_filter_is_none(self):
        assert _find_version({'filters': {'__version__': None}}) == None
        assert _find_version({'filters': {'__version__': [None]}}) == None
        assert _find_version({'filters': {'__version__': [None, None]}}) == None
        # only the first value is used if there is a list, even if the first value is invalid
        assert _find_version({'filters': {'__version__': [None, 10]}}) == None

    def test_version_as_filter_exists(self):
        assert _find_version({'filters': {'__version__': 10}}) == 10
        assert _find_version({'filters': {'__version__': '10'}}) == 10
        assert _find_version({'filters': {'__version__': [10]}}) == 10
        assert _find_version({'filters': {'__version__': ['10']}}) == 10
        # only the first value is used
        assert _find_version({'filters': {'__version__': ['10', None]}}) == 10

        with pytest.raises(ValueError):
            _find_version({'filters': {'__version__': 'aaaaa'}})

        with pytest.raises(ValueError):
            _find_version({'filters': {'__version__': ['aaaaa']}})

        with pytest.raises(ValueError):
            _find_version({'filters': {'__version__': ['aaaaa', None]}})

    def test_both_are_none(self):
        assert (
            _find_version({'version': None, 'filters': {'__version__': None}}) == None
        )
        assert (
            _find_version({'version': None, 'filters': {'__version__': [None]}}) == None
        )

    def test_version_takes_precedence(self):
        assert _find_version({'version': 10, 'filters': {'__version__': 12}}) == 10
        assert _find_version({'version': 10, 'filters': {'__version__': [12]}}) == 10

    def test_both_mix(self):
        assert _find_version({'version': 10, 'filters': {'__version__': None}}) == 10
        assert _find_version({'version': 10, 'filters': {'__version__': [None]}}) == 10
        assert _find_version({'version': None, 'filters': {'__version__': 12}}) == 12
        assert _find_version({'version': None, 'filters': {'__version__': [12]}}) == 12


class TestCreateSearch(object):
    @contextlib.contextmanager
    def _patch(self, target, replacement):
        with patch(
            f'ckanext.versioned_datastore.lib.basic_query.search.{target}', replacement
        ):
            yield

    def test_onward_calls(self):
        # the params
        context = MagicMock()
        data_dict = MagicMock()
        original_data_dict = MagicMock()

        # the returns of our various mocked functions
        mock_validate_return = MagicMock()
        build_mock_return = MagicMock()
        find_version_return = MagicMock()

        # the mock functions
        find_version_mock = MagicMock(return_value=find_version_return)
        build_search_object_mock = MagicMock(return_value=build_mock_return)
        mock_plugin = MagicMock(
            datastore_modify_data_dict=MagicMock(return_value=mock_validate_return),
            datastore_modify_search=MagicMock(return_value=build_mock_return),
        )
        impl_mock = MagicMock(return_value=[mock_plugin])
        with self._patch('PluginImplementations', impl_mock):
            with self._patch('_find_version', find_version_mock):
                with self._patch('build_search_object', build_search_object_mock):
                    result = create_search(context, data_dict, original_data_dict)

        # 4 values are returned
        assert len(result) == 4
        # check the values
        assert result[0] == original_data_dict
        assert result[1] == mock_validate_return
        assert result[2] == find_version_return
        assert result[3] == build_mock_return
        # check the onward calls
        assert mock_plugin.datastore_modify_data_dict.call_count == 1
        assert mock_plugin.datastore_modify_search.call_count == 1
        assert find_version_mock.call_count == 1
        assert build_search_object_mock.call_count == 1

    def test_simple_usage(self):
        # create our own schema so that we don't have to mock a resource existing to pass the
        # actual datastore search schema
        context = {'schema': {'q': [str], 'version': [int]}}
        data_dict = {'q': 'banana', 'version': 23}
        original_data_dict = {'q': 'banana', 'version': 23, 'something_else': 29}

        result = create_search(context, data_dict, original_data_dict)

        # 4 values are returned
        assert len(result) == 4
        # check the values
        assert result[0] == original_data_dict
        assert result[1] == data_dict
        assert result[2] == 23
        # we don't care in this test if the Search has been created correctly, only that we get a
        # Search object back
        assert isinstance(result[3], Search)


class TestBuildSearchObject(object):
    def _run_test(self, kwargs, expected_result, add_size=True, add_sort=True):
        if add_size:
            expected_result['size'] = 100
        if add_sort:
            expected_result['sort'] = [prefix_field('_id')]
        assert build_search_object(**kwargs).to_dict() == expected_result
        assert build_search_object(**kwargs) == Search().from_dict(expected_result)

    def test_blank(self):
        self._run_test({}, {})

    def test_q_empties(self):
        self._run_test({'q': ''}, {})
        self._run_test({'q': None}, {})
        self._run_test({'q': {}}, {})

    def test_q_simple_text(self):
        self._run_test(
            {'q': 'banana'},
            {
                'query': {
                    'match': {
                        'meta.all': {
                            'query': 'banana',
                            'operator': 'and',
                        }
                    }
                }
            },
        )
        self._run_test(
            {'q': 'a multi-word example'},
            {
                'query': {
                    'match': {
                        'meta.all': {
                            'query': 'a multi-word example',
                            'operator': 'and',
                        }
                    }
                }
            },
        )

    def test_q_dicts(self):
        self._run_test(
            {'q': {'': 'banana'}},
            {
                'query': {
                    'match': {
                        'meta.all': {
                            'query': 'banana',
                            'operator': 'and',
                        }
                    }
                }
            },
        )
        self._run_test(
            {'q': {'field1': 'banana'}},
            {
                'query': {
                    'match': {
                        prefix_field('field1'): {
                            'query': 'banana',
                            'operator': 'and',
                        }
                    }
                }
            },
        )
        self._run_test(
            {'q': {'field1': 'banana', 'field2': 'lemons'}},
            {
                'query': {
                    'bool': {
                        'must': [
                            {
                                'match': {
                                    prefix_field('field1'): {
                                        'query': 'banana',
                                        'operator': 'and',
                                    },
                                }
                            },
                            {
                                'match': {
                                    prefix_field('field2'): {
                                        'query': 'lemons',
                                        'operator': 'and',
                                    },
                                }
                            },
                        ]
                    }
                }
            },
        )

    def test_q_string_and_unicode(self):
        self._run_test(
            {'q': 'a string'},
            {
                'query': {
                    'match': {
                        'meta.all': {
                            'query': 'a string',
                            'operator': 'and',
                        }
                    }
                }
            },
        )
        self._run_test(
            {'q': 'a unicode string'},
            {
                'query': {
                    'match': {
                        'meta.all': {
                            'query': 'a unicode string',
                            'operator': 'and',
                        }
                    }
                }
            },
        )

    def test_q_non_string(self):
        self._run_test(
            {'q': 4},
            {
                'query': {
                    'match': {
                        'meta.all': {
                            'query': 4,
                            'operator': 'and',
                        }
                    }
                }
            },
        )
        self._run_test(
            {'q': 4.31932},
            {
                'query': {
                    'match': {
                        'meta.all': {
                            'query': 4.31932,
                            'operator': 'and',
                        }
                    }
                }
            },
        )

    def test_filters_empties(self):
        self._run_test({'filters': {}}, {})

    def test_filters_non_lists(self):
        self._run_test(
            {
                'filters': {
                    'field1': 'banana',
                }
            },
            {
                'query': {
                    'bool': {
                        'filter': [
                            {
                                'term': {
                                    prefix_field('field1'): 'banana',
                                }
                            }
                        ]
                    }
                }
            },
        )
        self._run_test(
            {
                'filters': {
                    'field1': 'banana',
                    'field2': 'lemons',
                }
            },
            {
                'query': {
                    'bool': {
                        'filter': [
                            {
                                'term': {
                                    prefix_field('field1'): 'banana',
                                }
                            },
                            {
                                'term': {
                                    prefix_field('field2'): 'lemons',
                                }
                            },
                        ]
                    }
                }
            },
        )

        add_geo_search_mock = MagicMock(side_effect=lambda s, v: s)
        mock_geo_value = MagicMock()
        with patch(
            'ckanext.versioned_datastore.lib.basic_query.search.add_geo_search',
            add_geo_search_mock,
        ):
            build_search_object(filters={'__geo__': mock_geo_value})
        assert add_geo_search_mock.call_count == 1
        search_object, filter_value = add_geo_search_mock.call_args[0]
        assert isinstance(search_object, Search)
        assert mock_geo_value == filter_value

    def test_filters_lists(self):
        self._run_test(
            {
                'filters': {
                    'field1': ['banana'],
                }
            },
            {
                'query': {
                    'bool': {
                        'filter': [
                            {
                                'term': {
                                    prefix_field('field1'): 'banana',
                                }
                            }
                        ]
                    }
                }
            },
        )
        self._run_test(
            {
                'filters': {
                    'field1': ['banana'],
                    'field2': ['lemons'],
                }
            },
            {
                'query': {
                    'bool': {
                        'filter': [
                            {
                                'term': {
                                    prefix_field('field1'): 'banana',
                                }
                            },
                            {
                                'term': {
                                    prefix_field('field2'): 'lemons',
                                }
                            },
                        ]
                    }
                }
            },
        )
        self._run_test(
            {
                'filters': {
                    'field1': ['banana', 'goat', 'funnel'],
                    'field2': ['lemons', 'chunk'],
                }
            },
            {
                'query': {
                    'bool': {
                        'filter': [
                            {
                                'term': {
                                    prefix_field('field1'): 'banana',
                                }
                            },
                            {
                                'term': {
                                    prefix_field('field1'): 'goat',
                                }
                            },
                            {
                                'term': {
                                    prefix_field('field1'): 'funnel',
                                }
                            },
                            {
                                'term': {
                                    prefix_field('field2'): 'lemons',
                                }
                            },
                            {
                                'term': {
                                    prefix_field('field2'): 'chunk',
                                }
                            },
                        ]
                    }
                }
            },
        )
        add_geo_search_mock = MagicMock(side_effect=lambda s, v: s)
        mock_geo_value = MagicMock()
        with patch(
            'ckanext.versioned_datastore.lib.basic_query.search.add_geo_search',
            add_geo_search_mock,
        ):
            build_search_object(filters={'__geo__': [mock_geo_value]})
        assert add_geo_search_mock.call_count == 1
        search_object, filter_value = add_geo_search_mock.call_args[0]
        assert isinstance(search_object, Search)
        assert mock_geo_value == filter_value

    def test_filters_mix(self):
        self._run_test(
            {
                'filters': {
                    'field1': 'banana',
                    'field2': ['lemons', 'blarp'],
                }
            },
            {
                'query': {
                    'bool': {
                        'filter': [
                            {
                                'term': {
                                    prefix_field('field1'): 'banana',
                                }
                            },
                            {
                                'term': {
                                    prefix_field('field2'): 'lemons',
                                }
                            },
                            {
                                'term': {
                                    prefix_field('field2'): 'blarp',
                                }
                            },
                        ]
                    }
                }
            },
        )

    def test_after(self):
        self._run_test({'after': 'some value'}, {'search_after': 'some value'})

    def test_offset(self):
        self._run_test({'offset': 12}, {'from': 12})
        self._run_test({'offset': '289'}, {'from': 289})

    def test_limit(self):
        self._run_test({'limit': 15}, {'size': 15}, add_size=False)
        self._run_test({'limit': '1500'}, {'size': 1500}, add_size=False)

    def test_fields(self):
        self._run_test(
            {'fields': ['field1', 'field2']},
            {'_source': ['data.field1', 'data.field2']},
        )

    def test_facets(self):
        self._run_test({'facets': []}, {})

        self._run_test(
            {'facets': ['field1']},
            {
                'aggs': {
                    'field1': {'terms': {'field': prefix_field('field1'), 'size': 10}}
                }
            },
        )

        self._run_test(
            {'facets': ['field1', 'field2', 'field3']},
            {
                'aggs': {
                    'field1': {'terms': {'field': prefix_field('field1'), 'size': 10}},
                    'field2': {'terms': {'field': prefix_field('field2'), 'size': 10}},
                    'field3': {'terms': {'field': prefix_field('field3'), 'size': 10}},
                }
            },
        )

        self._run_test(
            {
                'facets': ['field1', 'field2', 'field3'],
                'facet_limits': {
                    'field1': 20,
                    'field3': 100,
                },
            },
            {
                'aggs': {
                    'field1': {'terms': {'field': prefix_field('field1'), 'size': 20}},
                    'field2': {'terms': {'field': prefix_field('field2'), 'size': 10}},
                    'field3': {'terms': {'field': prefix_field('field3'), 'size': 100}},
                }
            },
        )

    def test_sorts(self):
        self._run_test({'sort': []}, {})
        self._run_test(
            {'sort': ['field1']},
            {'sort': [prefix_field('field1'), prefix_field('_id')]},
            add_sort=False,
        )
        self._run_test(
            {'sort': ['field1', 'field2']},
            {
                'sort': [
                    prefix_field('field1'),
                    prefix_field('field2'),
                    prefix_field('_id'),
                ]
            },
            add_sort=False,
        )
        self._run_test(
            {'sort': ['field1', '_id']},
            {'sort': [prefix_field('field1'), prefix_field('_id')]},
            add_sort=False,
        )
        self._run_test(
            {'sort': ['field1 desc', 'field2 asc']},
            {
                'sort': [
                    {prefix_field('field1'): {'order': 'desc'}},
                    prefix_field('field2'),
                    prefix_field('_id'),
                ]
            },
            add_sort=False,
        )
        self._run_test(
            {'sort': ['_id desc']},
            {'sort': [{prefix_field('_id'): {'order': 'desc'}}]},
            add_sort=False,
        )



class IngestionException(Exception):
    """
    Represents an exception that has occurred during ingestion.
    """

    pass


class UnsupportedDataSource(IngestionException):
    '''
    Should be raised when the data source we are attempting to ingest isn't one we can ingest - i.e.
    the format isn't one we support.
    '''

    def __init__(self, res_format):
        '''
        :param res_format: the resource format
        '''
        super(UnsupportedDataSource, self).__init__(
            f'Could not find ingest reader for {res_format if res_format else "n/a"}'
        )
        self.res_format = res_format


class InvalidId(IngestionException):
    """
    Should be raised when the data source we are attempting to ingest contains at least
    one row with an _id field which has a non-integer value.
    """

    def __init__(self, row_number, row, cause=None):
        '''
        :param row_number: the row number (1-indexed, excluding the header)
        :param row: the row (this should be a dict
        :param cause: optional cause exception, for example a ValueError thrown by int(row['_id')
        '''
        message = f'Row {row_number} had an invalid integer id: "{row["_id"]}"'
        if cause is not None:
            message = f'{message} [{cause.__class__.__name__}: {str(cause)}]'

        super(InvalidId, self).__init__(message)
        self.row_number = row_number
        self.row = row
        self.cause = cause


class DuplicateDataSource(IngestionException):
    """
    Should be raised when the data source we are attempting to ingest is the same as the
    last successful ingest's data source.
    """

    def __init__(self, file_hash):
        '''
        :param file_hash: the file hash that clashed
        '''
        super(DuplicateDataSource, self).__init__(
            f'This file has been ingested before, ignoring [hash: {file_hash}]'
        )
        self.file_hash = file_hash


class InvalidCharacterException(IngestionException):
    """
    Thrown when there is an invalid unicode character found in the resource data.

    This is detected
    by checking if the unicode version of the row contains any category C characters (control
    characters basically, see here: http://www.unicode.org/reports/tr44/#General_Category_Values).
    This is treated as an error to avoid putting crap unicode into the jsonl.gz intermediate file
    and then erroring when attempting to deserialise the json.
    Typically this error is produced when the user has uploaded a file in a really weird character
    encoding and we failed to detect it, thus falling back to UTF-8.
    """

    def __init__(self, row_number, row):
        '''
        :param row_number: the row number (1-indexed, excluding the header)
        :param row: the row (this should be a dict
        '''
        message = f'Row {row_number} (excluding header) contained an invalid character'
        super(InvalidCharacterException, self).__init__(message)
        self.row_number = row_number
        self.row = row
